"""
altgraph.ObjectGraph - Graph of objects with an identifier
==========================================================

A graph of objects that have a "graphident" attribute.
graphident is the key for the object in the graph
"""

from altgraph import GraphError 
from altgraph .Graph import Graph 
from altgraph .GraphUtil import filter_stack 


class ObjectGraph (object ):
    """
    A graph of objects that have a "graphident" attribute.
    graphident is the key for the object in the graph
    """

    def __init__ (self ,graph =None ,debug =0 ):
        if graph is None :
            graph =Graph ()
        self .graphident =self 
        self .graph =graph 
        self .debug =debug 
        self .indent =0 
        graph .add_node (self ,None )

    def __repr__ (self ):
        return "<%s>"%(type (self ).__name__ ,)

    def flatten (self ,condition =None ,start =None ):
        """
        Iterate over the subgraph that is entirely reachable by condition
        starting from the given start node or the ObjectGraph root
        """
        if start is None :
            start =self 
        start =self .getRawIdent (start )
        return self .graph .iterdata (start =start ,condition =condition )

    def nodes (self ):
        for ident in self .graph :
            node =self .graph .node_data (ident )
            if node is not None :
                yield self .graph .node_data (ident )

    def get_edges (self ,node ):
        if node is None :
            node =self 
        start =self .getRawIdent (node )
        _ ,_ ,outraw ,incraw =self .graph .describe_node (start )

        def iter_edges (lst ,n ):
            seen =set ()
            for tpl in (self .graph .describe_edge (e )for e in lst ):
                ident =tpl [n ]
                if ident not in seen :
                    yield self .findNode (ident )
                    seen .add (ident )

        return iter_edges (outraw ,3 ),iter_edges (incraw ,2 )

    def edgeData (self ,fromNode ,toNode ):
        if fromNode is None :
            fromNode =self 
        start =self .getRawIdent (fromNode )
        stop =self .getRawIdent (toNode )
        edge =self .graph .edge_by_node (start ,stop )
        return self .graph .edge_data (edge )

    def updateEdgeData (self ,fromNode ,toNode ,edgeData ):
        if fromNode is None :
            fromNode =self 
        start =self .getRawIdent (fromNode )
        stop =self .getRawIdent (toNode )
        edge =self .graph .edge_by_node (start ,stop )
        self .graph .update_edge_data (edge ,edgeData )

    def filterStack (self ,filters ):
        """
        Filter the ObjectGraph in-place by removing all edges to nodes that
        do not match every filter in the given filter list

        Returns a tuple containing the number of:
            (nodes_visited, nodes_removed, nodes_orphaned)
        """
        visited ,removes ,orphans =filter_stack (self .graph ,self ,filters )

        for last_good ,tail in orphans :
            self .graph .add_edge (last_good ,tail ,edge_data ="orphan")

        for node in removes :
            self .graph .hide_node (node )

        return len (visited )-1 ,len (removes ),len (orphans )

    def removeNode (self ,node ):
        """
        Remove the given node from the graph if it exists
        """
        ident =self .getIdent (node )
        if ident is not None :
            self .graph .hide_node (ident )

    def removeReference (self ,fromnode ,tonode ):
        """
        Remove all edges from fromnode to tonode
        """
        if fromnode is None :
            fromnode =self 
        fromident =self .getIdent (fromnode )
        toident =self .getIdent (tonode )
        if fromident is not None and toident is not None :
            while True :
                edge =self .graph .edge_by_node (fromident ,toident )
                if edge is None :
                    break 
                self .graph .hide_edge (edge )

    def getIdent (self ,node ):
        """
        Get the graph identifier for a node
        """
        ident =self .getRawIdent (node )
        if ident is not None :
            return ident 
        node =self .findNode (node )
        if node is None :
            return None 
        return node .graphident 

    def getRawIdent (self ,node ):
        """
        Get the identifier for a node object
        """
        if node is self :
            return node 
        ident =getattr (node ,"graphident",None )
        return ident 

    def __contains__ (self ,node ):
        return self .findNode (node )is not None 

    def findNode (self ,node ):
        """
        Find the node on the graph
        """
        ident =self .getRawIdent (node )
        if ident is None :
            ident =node 
        try :
            return self .graph .node_data (ident )
        except KeyError :
            return None 

    def addNode (self ,node ):
        """
        Add a node to the graph referenced by the root
        """
        self .msg (4 ,"addNode",node )

        try :
            self .graph .restore_node (node .graphident )
        except GraphError :
            self .graph .add_node (node .graphident ,node )

    def createReference (self ,fromnode ,tonode ,edge_data =None ):
        """
        Create a reference from fromnode to tonode
        """
        if fromnode is None :
            fromnode =self 
        fromident ,toident =self .getIdent (fromnode ),self .getIdent (tonode )
        if fromident is None or toident is None :
            return 
        self .msg (4 ,"createReference",fromnode ,tonode ,edge_data )
        self .graph .add_edge (fromident ,toident ,edge_data =edge_data )

    def createNode (self ,cls ,name ,*args ,**kw ):
        """
        Add a node of type cls to the graph if it does not already exist
        by the given name
        """
        m =self .findNode (name )
        if m is None :
            m =cls (name ,*args ,**kw )
            self .addNode (m )
        return m 

    def msg (self ,level ,s ,*args ):
        """
        Print a debug message with the given level
        """
        if s and level <=self .debug :
            print ("%s%s %s"%("  "*self .indent ,s ," ".join (map (repr ,args ))))

    def msgin (self ,level ,s ,*args ):
        """
        Print a debug message and indent
        """
        if level <=self .debug :
            self .msg (level ,s ,*args )
            self .indent =self .indent +1 

    def msgout (self ,level ,s ,*args ):
        """
        Dedent and print a debug message
        """
        if level <=self .debug :
            self .indent =self .indent -1 
            self .msg (level ,s ,*args )



def check (cmd ,mf ):
    m =mf .findNode ("sphinx")
    if m is None or m .filename is None :
        return None 

    includes =[
    "sphinxcontrib.applehelp",
    "sphinxcontrib.devhelp",
    "sphinxcontrib.htmlhelp",
    "sphinxcontrib.jsmath",
    "sphinxcontrib.qthelp",
    "sphinxcontrib.serializinghtml",
    ]

    return {"includes":includes }



"""
Other than changing the load commands in such a way that they do not
contain the load command itself, this is largely a by-hand conversion
of the C headers.  Hopefully everything in here should be at least as
obvious as the C headers, and you should be using the C headers as a real
reference because the documentation didn't come along for the ride.

Doing much of anything with the symbol tables or segments is really
not covered at this point.

See /usr/include/mach-o and friends.
"""

import time 

from macholib .ptypes import (
Structure ,
p_int32 ,
p_int64 ,
p_long ,
p_short ,
p_uint8 ,
p_uint32 ,
p_uint64 ,
p_ulong ,
pypackable ,
)

_CPU_ARCH_ABI64 =0x01000000 

CPU_TYPE_NAMES ={
-1 :"ANY",
1 :"VAX",
6 :"MC680x0",
7 :"i386",
_CPU_ARCH_ABI64 |7 :"x86_64",
8 :"MIPS",
10 :"MC98000",
11 :"HPPA",
12 :"ARM",
_CPU_ARCH_ABI64 |12 :"ARM64",
13 :"MC88000",
14 :"SPARC",
15 :"i860",
16 :"Alpha",
18 :"PowerPC",
_CPU_ARCH_ABI64 |18 :"PowerPC64",
}

INTEL64_SUBTYPE ={
3 :"CPU_SUBTYPE_X86_64_ALL",
4 :"CPU_SUBTYPE_X86_ARCH1",
8 :"CPU_SUBTYPE_X86_64_H",
}


INTEL_SUBTYPE ={
0 :"CPU_SUBTYPE_INTEL_MODEL_ALL",
1 :"CPU_THREADTYPE_INTEL_HTT",
3 :"CPU_SUBTYPE_I386_ALL",
4 :"CPU_SUBTYPE_486",
5 :"CPU_SUBTYPE_586",
8 :"CPU_SUBTYPE_PENTIUM_3",
9 :"CPU_SUBTYPE_PENTIUM_M",
10 :"CPU_SUBTYPE_PENTIUM_4",
11 :"CPU_SUBTYPE_ITANIUM",
12 :"CPU_SUBTYPE_XEON",
34 :"CPU_SUBTYPE_XEON_MP",
42 :"CPU_SUBTYPE_PENTIUM_4_M",
43 :"CPU_SUBTYPE_ITANIUM_2",
38 :"CPU_SUBTYPE_PENTPRO",
40 :"CPU_SUBTYPE_PENTIUM_3_M",
52 :"CPU_SUBTYPE_PENTIUM_3_XEON",
102 :"CPU_SUBTYPE_PENTII_M3",
132 :"CPU_SUBTYPE_486SX",
166 :"CPU_SUBTYPE_PENTII_M5",
199 :"CPU_SUBTYPE_CELERON",
231 :"CPU_SUBTYPE_CELERON_MOBILE",
}

MC680_SUBTYPE ={
1 :"CPU_SUBTYPE_MC680x0_ALL",
2 :"CPU_SUBTYPE_MC68040",
3 :"CPU_SUBTYPE_MC68030_ONLY",
}

MIPS_SUBTYPE ={
0 :"CPU_SUBTYPE_MIPS_ALL",
1 :"CPU_SUBTYPE_MIPS_R2300",
2 :"CPU_SUBTYPE_MIPS_R2600",
3 :"CPU_SUBTYPE_MIPS_R2800",
4 :"CPU_SUBTYPE_MIPS_R2000a",
5 :"CPU_SUBTYPE_MIPS_R2000",
6 :"CPU_SUBTYPE_MIPS_R3000a",
7 :"CPU_SUBTYPE_MIPS_R3000",
}

MC98000_SUBTYPE ={0 :"CPU_SUBTYPE_MC98000_ALL",1 :"CPU_SUBTYPE_MC98601"}

HPPA_SUBTYPE ={0 :"CPU_SUBTYPE_HPPA_7100",1 :"CPU_SUBTYPE_HPPA_7100LC"}

MC88_SUBTYPE ={
0 :"CPU_SUBTYPE_MC88000_ALL",
1 :"CPU_SUBTYPE_MC88100",
2 :"CPU_SUBTYPE_MC88110",
}

SPARC_SUBTYPE ={0 :"CPU_SUBTYPE_SPARC_ALL"}

I860_SUBTYPE ={0 :"CPU_SUBTYPE_I860_ALL",1 :"CPU_SUBTYPE_I860_860"}

POWERPC_SUBTYPE ={
0 :"CPU_SUBTYPE_POWERPC_ALL",
1 :"CPU_SUBTYPE_POWERPC_601",
2 :"CPU_SUBTYPE_POWERPC_602",
3 :"CPU_SUBTYPE_POWERPC_603",
4 :"CPU_SUBTYPE_POWERPC_603e",
5 :"CPU_SUBTYPE_POWERPC_603ev",
6 :"CPU_SUBTYPE_POWERPC_604",
7 :"CPU_SUBTYPE_POWERPC_604e",
8 :"CPU_SUBTYPE_POWERPC_620",
9 :"CPU_SUBTYPE_POWERPC_750",
10 :"CPU_SUBTYPE_POWERPC_7400",
11 :"CPU_SUBTYPE_POWERPC_7450",
100 :"CPU_SUBTYPE_POWERPC_970",
}

ARM_SUBTYPE ={
0 :"CPU_SUBTYPE_ARM_ALL12",
5 :"CPU_SUBTYPE_ARM_V4T",
6 :"CPU_SUBTYPE_ARM_V6",
7 :"CPU_SUBTYPE_ARM_V5TEJ",
8 :"CPU_SUBTYPE_ARM_XSCALE",
9 :"CPU_SUBTYPE_ARM_V7",
10 :"CPU_SUBTYPE_ARM_V7F",
11 :"CPU_SUBTYPE_ARM_V7S",
12 :"CPU_SUBTYPE_ARM_V7K",
13 :"CPU_SUBTYPE_ARM_V8",
14 :"CPU_SUBTYPE_ARM_V6M",
15 :"CPU_SUBTYPE_ARM_V7M",
16 :"CPU_SUBTYPE_ARM_V7EM",
17 :"CPU_SUBTYPE_ARM_V8M",
}

ARM64_SUBTYPE ={
0 :"CPU_SUBTYPE_ARM64_ALL",
1 :"CPU_SUBTYPE_ARM64_V8",
2 :"CPU_SUBTYPE_ARM64E",
}

VAX_SUBTYPE ={
0 :"CPU_SUBTYPE_VAX_ALL",
1 :"CPU_SUBTYPE_VAX780",
2 :"CPU_SUBTYPE_VAX785",
3 :"CPU_SUBTYPE_VAX750",
4 :"CPU_SUBTYPE_VAX730",
5 :"CPU_SUBTYPE_UVAXI",
6 :"CPU_SUBTYPE_UVAXII",
7 :"CPU_SUBTYPE_VAX8200",
8 :"CPU_SUBTYPE_VAX8500",
9 :"CPU_SUBTYPE_VAX8600",
10 :"CPU_SUBTYPE_VAX8650",
11 :"CPU_SUBTYPE_VAX8800",
12 :"CPU_SUBTYPE_UVAXIII",
}


def get_cpu_subtype (cpu_type ,cpu_subtype ):
    st =cpu_subtype &0x0FFFFFFF 

    if cpu_type ==1 :
        subtype =VAX_SUBTYPE .get (st ,st )
    elif cpu_type ==6 :
        subtype =MC680_SUBTYPE .get (st ,st )
    elif cpu_type ==7 :
        subtype =INTEL_SUBTYPE .get (st ,st )
    elif cpu_type ==7 |_CPU_ARCH_ABI64 :
        subtype =INTEL64_SUBTYPE .get (st ,st )
    elif cpu_type ==8 :
        subtype =MIPS_SUBTYPE .get (st ,st )
    elif cpu_type ==10 :
        subtype =MC98000_SUBTYPE .get (st ,st )
    elif cpu_type ==11 :
        subtype =HPPA_SUBTYPE .get (st ,st )
    elif cpu_type ==12 :
        subtype =ARM_SUBTYPE .get (st ,st )
    elif cpu_type ==12 |_CPU_ARCH_ABI64 :
        subtype =ARM64_SUBTYPE .get (st ,st )
    elif cpu_type ==13 :
        subtype =MC88_SUBTYPE .get (st ,st )
    elif cpu_type ==14 :
        subtype =SPARC_SUBTYPE .get (st ,st )
    elif cpu_type ==15 :
        subtype =I860_SUBTYPE .get (st ,st )
    elif cpu_type ==16 :
        subtype =MIPS_SUBTYPE .get (st ,st )
    elif cpu_type ==18 :
        subtype =POWERPC_SUBTYPE .get (st ,st )
    elif cpu_type ==18 |_CPU_ARCH_ABI64 :
        subtype =POWERPC_SUBTYPE .get (st ,st )
    else :
        subtype =str (st )

    return subtype 


_MH_EXECUTE_SYM ="__mh_execute_header"
MH_EXECUTE_SYM ="_mh_execute_header"
_MH_BUNDLE_SYM ="__mh_bundle_header"
MH_BUNDLE_SYM ="_mh_bundle_header"
_MH_DYLIB_SYM ="__mh_dylib_header"
MH_DYLIB_SYM ="_mh_dylib_header"
_MH_DYLINKER_SYM ="__mh_dylinker_header"
MH_DYLINKER_SYM ="_mh_dylinker_header"

(
MH_OBJECT ,
MH_EXECUTE ,
MH_FVMLIB ,
MH_CORE ,
MH_PRELOAD ,
MH_DYLIB ,
MH_DYLINKER ,
MH_BUNDLE ,
MH_DYLIB_STUB ,
MH_DSYM ,
)=range (0x1 ,0xB )

MH_FILESET =0xC 

(
MH_NOUNDEFS ,
MH_INCRLINK ,
MH_DYLDLINK ,
MH_BINDATLOAD ,
MH_PREBOUND ,
MH_SPLIT_SEGS ,
MH_LAZY_INIT ,
MH_TWOLEVEL ,
MH_FORCE_FLAT ,
MH_NOMULTIDEFS ,
MH_NOFIXPREBINDING ,
MH_PREBINDABLE ,
MH_ALLMODSBOUND ,
MH_SUBSECTIONS_VIA_SYMBOLS ,
MH_CANONICAL ,
MH_WEAK_DEFINES ,
MH_BINDS_TO_WEAK ,
MH_ALLOW_STACK_EXECUTION ,
MH_ROOT_SAFE ,
MH_SETUID_SAFE ,
MH_NO_REEXPORTED_DYLIBS ,
MH_PIE ,
MH_DEAD_STRIPPABLE_DYLIB ,
MH_HAS_TLV_DESCRIPTORS ,
MH_NO_HEAP_EXECUTION ,
MH_APP_EXTENSION_SAFE ,
)=map ((1 ).__lshift__ ,range (26 ))

MH_MAGIC =0xFEEDFACE 
MH_CIGAM =0xCEFAEDFE 
MH_MAGIC_64 =0xFEEDFACF 
MH_CIGAM_64 =0xCFFAEDFE 

integer_t =p_int32 
cpu_type_t =integer_t 
cpu_subtype_t =p_uint32 

MH_FILETYPE_NAMES ={
MH_OBJECT :"relocatable object",
MH_EXECUTE :"demand paged executable",
MH_FVMLIB :"fixed vm shared library",
MH_CORE :"core",
MH_PRELOAD :"preloaded executable",
MH_DYLIB :"dynamically bound shared library",
MH_DYLINKER :"dynamic link editor",
MH_BUNDLE :"dynamically bound bundle",
MH_DYLIB_STUB :"shared library stub for static linking",
MH_DSYM :"symbol information",
MH_FILESET :"fileset object",
}

MH_FILETYPE_SHORTNAMES ={
MH_OBJECT :"object",
MH_EXECUTE :"execute",
MH_FVMLIB :"fvmlib",
MH_CORE :"core",
MH_PRELOAD :"preload",
MH_DYLIB :"dylib",
MH_DYLINKER :"dylinker",
MH_BUNDLE :"bundle",
MH_DYLIB_STUB :"dylib_stub",
MH_DSYM :"dsym",
}

MH_FLAGS_NAMES ={
MH_NOUNDEFS :"MH_NOUNDEFS",
MH_INCRLINK :"MH_INCRLINK",
MH_DYLDLINK :"MH_DYLDLINK",
MH_BINDATLOAD :"MH_BINDATLOAD",
MH_PREBOUND :"MH_PREBOUND",
MH_SPLIT_SEGS :"MH_SPLIT_SEGS",
MH_LAZY_INIT :"MH_LAZY_INIT",
MH_TWOLEVEL :"MH_TWOLEVEL",
MH_FORCE_FLAT :"MH_FORCE_FLAT",
MH_NOMULTIDEFS :"MH_NOMULTIDEFS",
MH_NOFIXPREBINDING :"MH_NOFIXPREBINDING",
MH_PREBINDABLE :"MH_PREBINDABLE",
MH_ALLMODSBOUND :"MH_ALLMODSBOUND",
MH_SUBSECTIONS_VIA_SYMBOLS :"MH_SUBSECTIONS_VIA_SYMBOLS",
MH_CANONICAL :"MH_CANONICAL",
MH_WEAK_DEFINES :"MH_WEAK_DEFINES",
MH_BINDS_TO_WEAK :"MH_BINDS_TO_WEAK",
MH_ALLOW_STACK_EXECUTION :"MH_ALLOW_STACK_EXECUTION",
MH_ROOT_SAFE :"MH_ROOT_SAFE",
MH_SETUID_SAFE :"MH_SETUID_SAFE",
MH_NO_REEXPORTED_DYLIBS :"MH_NO_REEXPORTED_DYLIBS",
MH_PIE :"MH_PIE",
MH_DEAD_STRIPPABLE_DYLIB :"MH_DEAD_STRIPPABLE_DYLIB",
MH_HAS_TLV_DESCRIPTORS :"MH_HAS_TLV_DESCRIPTORS",
MH_NO_HEAP_EXECUTION :"MH_NO_HEAP_EXECUTION",
MH_APP_EXTENSION_SAFE :"MH_APP_EXTENSION_SAFE",
}

MH_FLAGS_DESCRIPTIONS ={
MH_NOUNDEFS :"no undefined references",
MH_INCRLINK :"output of an incremental link",
MH_DYLDLINK :"input for the dynamic linker",
MH_BINDATLOAD :"undefined references bound dynamically when loaded",
MH_PREBOUND :"dynamic undefined references prebound",
MH_SPLIT_SEGS :"split read-only and read-write segments",
MH_LAZY_INIT :"(obsolete)",
MH_TWOLEVEL :"using two-level name space bindings",
MH_FORCE_FLAT :"forcing all imagges to use flat name space bindings",
MH_NOMULTIDEFS :"umbrella guarantees no multiple definitions",
MH_NOFIXPREBINDING :"do not notify prebinding agent about this executable",
MH_PREBINDABLE :"the binary is not prebound but can have its prebinding redone",
MH_ALLMODSBOUND :"indicates that this binary binds to all "
"two-level namespace modules of its dependent libraries",
MH_SUBSECTIONS_VIA_SYMBOLS :"safe to divide up the sections into "
"sub-sections via symbols for dead code stripping",
MH_CANONICAL :"the binary has been canonicalized via the unprebind operation",
MH_WEAK_DEFINES :"the final linked image contains external weak symbols",
MH_BINDS_TO_WEAK :"the final linked image uses weak symbols",
MH_ALLOW_STACK_EXECUTION :"all stacks in the task will be given "
"stack execution privilege",
MH_ROOT_SAFE :"the binary declares it is safe for use in processes with uid zero",
MH_SETUID_SAFE :"the binary declares it is safe for use in processes "
"when issetugid() is true",
MH_NO_REEXPORTED_DYLIBS :"the static linker does not need to examine dependent "
"dylibs to see if any are re-exported",
MH_PIE :"the OS will load the main executable at a random address",
MH_DEAD_STRIPPABLE_DYLIB :"the static linker will automatically not create a "
"LC_LOAD_DYLIB load command to the dylib if no symbols are being "
"referenced from the dylib",
MH_HAS_TLV_DESCRIPTORS :"contains a section of type S_THREAD_LOCAL_VARIABLES",
MH_NO_HEAP_EXECUTION :"the OS will run the main executable with a "
"non-executable heap even on platforms that don't require it",
MH_APP_EXTENSION_SAFE :"the code was linked for use in an application extension.",
}


class mach_version_helper (Structure ):
    _fields_ =(("_version",p_uint32 ),)

    @property 
    def major (self ):
        return self ._version >>16 &0xFFFF 

    @major .setter 
    def major (self ,v ):
        self ._version =(self ._version &0xFFFF )|(v <<16 )

    @property 
    def minor (self ):
        return self ._version >>8 &0xFF 

    @minor .setter 
    def minor (self ,v ):
        self ._version =(self ._version &0xFFFF00FF )|(v <<8 )

    @property 
    def rev (self ):
        return self ._version &0xFF 

    @rev .setter 
    def rev (self ,v ):
        return (self ._version &0xFFFFFF00 )|v 

    def __str__ (self ):
        return "%s.%s.%s"%(self .major ,self .minor ,self .rev )


class mach_timestamp_helper (p_uint32 ):
    def __str__ (self ):
        return time .ctime (self )


def read_struct (f ,s ,**kw ):
    return s .from_fileobj (f ,**kw )


class mach_header (Structure ):
    _fields_ =(
    ("magic",p_uint32 ),
    ("cputype",cpu_type_t ),
    ("cpusubtype",cpu_subtype_t ),
    ("filetype",p_uint32 ),
    ("ncmds",p_uint32 ),
    ("sizeofcmds",p_uint32 ),
    ("flags",p_uint32 ),
    )

    def _describe (self ):
        bit =1 
        flags =self .flags 
        dflags =[]
        while flags and bit <(1 <<32 ):
            if flags &bit :
                dflags .append (
                {
                "name":MH_FLAGS_NAMES .get (bit ,str (bit )),
                "description":MH_FLAGS_DESCRIPTIONS .get (bit ,str (bit )),
                }
                )
                flags =flags ^bit 
            bit <<=1 
        return (
        ("magic",int (self .magic )),
        ("cputype_string",CPU_TYPE_NAMES .get (self .cputype ,self .cputype )),
        ("cputype",int (self .cputype )),
        ("cpusubtype_string",get_cpu_subtype (self .cputype ,self .cpusubtype )),
        ("cpusubtype",int (self .cpusubtype )),
        ("filetype_string",MH_FILETYPE_NAMES .get (self .filetype ,self .filetype )),
        ("filetype",int (self .filetype )),
        ("ncmds",self .ncmds ),
        ("sizeofcmds",self .sizeofcmds ),
        ("flags",dflags ),
        ("raw_flags",int (self .flags )),
        )


class mach_header_64 (mach_header ):
    _fields_ =mach_header ._fields_ +(("reserved",p_uint32 ),)


class load_command (Structure ):
    _fields_ =(("cmd",p_uint32 ),("cmdsize",p_uint32 ))

    def get_cmd_name (self ):
        return LC_NAMES .get (self .cmd ,self .cmd )


LC_REQ_DYLD =0x80000000 

(
LC_SEGMENT ,
LC_SYMTAB ,
LC_SYMSEG ,
LC_THREAD ,
LC_UNIXTHREAD ,
LC_LOADFVMLIB ,
LC_IDFVMLIB ,
LC_IDENT ,
LC_FVMFILE ,
LC_PREPAGE ,
LC_DYSYMTAB ,
LC_LOAD_DYLIB ,
LC_ID_DYLIB ,
LC_LOAD_DYLINKER ,
LC_ID_DYLINKER ,
LC_PREBOUND_DYLIB ,
LC_ROUTINES ,
LC_SUB_FRAMEWORK ,
LC_SUB_UMBRELLA ,
LC_SUB_CLIENT ,
LC_SUB_LIBRARY ,
LC_TWOLEVEL_HINTS ,
LC_PREBIND_CKSUM ,
)=range (0x1 ,0x18 )

LC_LOAD_WEAK_DYLIB =LC_REQ_DYLD |0x18 

LC_SEGMENT_64 =0x19 
LC_ROUTINES_64 =0x1A 
LC_UUID =0x1B 
LC_RPATH =0x1C |LC_REQ_DYLD 
LC_CODE_SIGNATURE =0x1D 
LC_CODE_SEGMENT_SPLIT_INFO =0x1E 
LC_REEXPORT_DYLIB =0x1F |LC_REQ_DYLD 
LC_LAZY_LOAD_DYLIB =0x20 
LC_ENCRYPTION_INFO =0x21 
LC_DYLD_INFO =0x22 
LC_DYLD_INFO_ONLY =0x22 |LC_REQ_DYLD 
LC_LOAD_UPWARD_DYLIB =0x23 |LC_REQ_DYLD 
LC_VERSION_MIN_MACOSX =0x24 
LC_VERSION_MIN_IPHONEOS =0x25 
LC_FUNCTION_STARTS =0x26 
LC_DYLD_ENVIRONMENT =0x27 
LC_MAIN =0x28 |LC_REQ_DYLD 
LC_DATA_IN_CODE =0x29 
LC_SOURCE_VERSION =0x2A 
LC_DYLIB_CODE_SIGN_DRS =0x2B 
LC_ENCRYPTION_INFO_64 =0x2C 
LC_LINKER_OPTION =0x2D 
LC_LINKER_OPTIMIZATION_HINT =0x2E 
LC_VERSION_MIN_TVOS =0x2F 
LC_VERSION_MIN_WATCHOS =0x30 
LC_NOTE =0x31 
LC_BUILD_VERSION =0x32 
LC_DYLD_EXPORTS_TRIE =0x33 |LC_REQ_DYLD 
LC_DYLD_CHAINED_FIXUPS =0x34 |LC_REQ_DYLD 
LC_FILESET_ENTRY =0x35 |LC_REQ_DYLD 



class lc_str (p_uint32 ):
    pass 


p_str16 =pypackable ("p_str16",bytes ,"16s")

vm_prot_t =p_int32 


class segment_command (Structure ):
    _fields_ =(
    ("segname",p_str16 ),
    ("vmaddr",p_uint32 ),
    ("vmsize",p_uint32 ),
    ("fileoff",p_uint32 ),
    ("filesize",p_uint32 ),
    ("maxprot",vm_prot_t ),
    ("initprot",vm_prot_t ),
    ("nsects",p_uint32 ),
    ("flags",p_uint32 ),
    )

    def describe (self ):
        s ={}
        s ["segname"]=self .segname .rstrip ("\x00")
        s ["vmaddr"]=int (self .vmaddr )
        s ["vmsize"]=int (self .vmsize )
        s ["fileoff"]=int (self .fileoff )
        s ["filesize"]=int (self .filesize )
        s ["initprot"]=self .get_initial_virtual_memory_protections ()
        s ["initprot_raw"]=int (self .initprot )
        s ["maxprot"]=self .get_max_virtual_memory_protections ()
        s ["maxprot_raw"]=int (self .maxprot )
        s ["nsects"]=int (self .nsects )
        s ["flags"]=self .flags 
        return s 

    def get_initial_virtual_memory_protections (self ):
        vm =[]
        if self .initprot ==0 :
            vm .append ("VM_PROT_NONE")
        if self .initprot &1 :
            vm .append ("VM_PROT_READ")
        if self .initprot &2 :
            vm .append ("VM_PROT_WRITE")
        if self .initprot &4 :
            vm .append ("VM_PROT_EXECUTE")
        return vm 

    def get_max_virtual_memory_protections (self ):
        vm =[]
        if self .maxprot ==0 :
            vm .append ("VM_PROT_NONE")
        if self .maxprot &1 :
            vm .append ("VM_PROT_READ")
        if self .maxprot &2 :
            vm .append ("VM_PROT_WRITE")
        if self .maxprot &4 :
            vm .append ("VM_PROT_EXECUTE")
        return vm 


class segment_command_64 (Structure ):
    _fields_ =(
    ("segname",p_str16 ),
    ("vmaddr",p_uint64 ),
    ("vmsize",p_uint64 ),
    ("fileoff",p_uint64 ),
    ("filesize",p_uint64 ),
    ("maxprot",vm_prot_t ),
    ("initprot",vm_prot_t ),
    ("nsects",p_uint32 ),
    ("flags",p_uint32 ),
    )

    def describe (self ):
        s ={}
        s ["segname"]=self .segname .rstrip ("\x00")
        s ["vmaddr"]=int (self .vmaddr )
        s ["vmsize"]=int (self .vmsize )
        s ["fileoff"]=int (self .fileoff )
        s ["filesize"]=int (self .filesize )
        s ["initprot"]=self .get_initial_virtual_memory_protections ()
        s ["initprot_raw"]=int (self .initprot )
        s ["maxprot"]=self .get_max_virtual_memory_protections ()
        s ["maxprot_raw"]=int (self .maxprot )
        s ["nsects"]=int (self .nsects )
        s ["flags"]=self .flags 
        return s 

    def get_initial_virtual_memory_protections (self ):
        vm =[]
        if self .initprot ==0 :
            vm .append ("VM_PROT_NONE")
        if self .initprot &1 :
            vm .append ("VM_PROT_READ")
        if self .initprot &2 :
            vm .append ("VM_PROT_WRITE")
        if self .initprot &4 :
            vm .append ("VM_PROT_EXECUTE")
        return vm 

    def get_max_virtual_memory_protections (self ):
        vm =[]
        if self .maxprot ==0 :
            vm .append ("VM_PROT_NONE")
        if self .maxprot &1 :
            vm .append ("VM_PROT_READ")
        if self .maxprot &2 :
            vm .append ("VM_PROT_WRITE")
        if self .maxprot &4 :
            vm .append ("VM_PROT_EXECUTE")
        return vm 


SG_HIGHVM =0x1 
SG_FVMLIB =0x2 
SG_NORELOC =0x4 
SG_PROTECTED_VERSION_1 =0x8 


class section (Structure ):
    _fields_ =(
    ("sectname",p_str16 ),
    ("segname",p_str16 ),
    ("addr",p_uint32 ),
    ("size",p_uint32 ),
    ("offset",p_uint32 ),
    ("align",p_uint32 ),
    ("reloff",p_uint32 ),
    ("nreloc",p_uint32 ),
    ("flags",p_uint32 ),
    ("reserved1",p_uint32 ),
    ("reserved2",p_uint32 ),
    )

    def describe (self ):
        s ={}
        s ["sectname"]=self .sectname .rstrip ("\x00")
        s ["segname"]=self .segname .rstrip ("\x00")
        s ["addr"]=int (self .addr )
        s ["size"]=int (self .size )
        s ["offset"]=int (self .offset )
        s ["align"]=int (self .align )
        s ["reloff"]=int (self .reloff )
        s ["nreloc"]=int (self .nreloc )
        f ={}
        f ["type"]=FLAG_SECTION_TYPES [int (self .flags )&0xFF ]
        f ["attributes"]=[]
        for k in FLAG_SECTION_ATTRIBUTES :
            if k &self .flags :
                f ["attributes"].append (FLAG_SECTION_ATTRIBUTES [k ])
        if not f ["attributes"]:
            del f ["attributes"]
        s ["flags"]=f 
        s ["reserved1"]=int (self .reserved1 )
        s ["reserved2"]=int (self .reserved2 )
        return s 

    def add_section_data (self ,data ):
        self .section_data =data 


class section_64 (Structure ):
    _fields_ =(
    ("sectname",p_str16 ),
    ("segname",p_str16 ),
    ("addr",p_uint64 ),
    ("size",p_uint64 ),
    ("offset",p_uint32 ),
    ("align",p_uint32 ),
    ("reloff",p_uint32 ),
    ("nreloc",p_uint32 ),
    ("flags",p_uint32 ),
    ("reserved1",p_uint32 ),
    ("reserved2",p_uint32 ),
    ("reserved3",p_uint32 ),
    )

    def describe (self ):
        s ={}
        s ["sectname"]=self .sectname .rstrip ("\x00")
        s ["segname"]=self .segname .rstrip ("\x00")
        s ["addr"]=int (self .addr )
        s ["size"]=int (self .size )
        s ["offset"]=int (self .offset )
        s ["align"]=int (self .align )
        s ["reloff"]=int (self .reloff )
        s ["nreloc"]=int (self .nreloc )
        f ={}
        f ["type"]=FLAG_SECTION_TYPES [int (self .flags )&0xFF ]
        f ["attributes"]=[]
        for k in FLAG_SECTION_ATTRIBUTES :
            if k &self .flags :
                f ["attributes"].append (FLAG_SECTION_ATTRIBUTES [k ])
        if not f ["attributes"]:
            del f ["attributes"]
        s ["flags"]=f 
        s ["reserved1"]=int (self .reserved1 )
        s ["reserved2"]=int (self .reserved2 )
        s ["reserved3"]=int (self .reserved3 )
        return s 

    def add_section_data (self ,data ):
        self .section_data =data 


SECTION_TYPE =0xFF 
SECTION_ATTRIBUTES =0xFFFFFF00 
S_REGULAR =0x0 
S_ZEROFILL =0x1 
S_CSTRING_LITERALS =0x2 
S_4BYTE_LITERALS =0x3 
S_8BYTE_LITERALS =0x4 
S_LITERAL_POINTERS =0x5 
S_NON_LAZY_SYMBOL_POINTERS =0x6 
S_LAZY_SYMBOL_POINTERS =0x7 
S_SYMBOL_STUBS =0x8 
S_MOD_INIT_FUNC_POINTERS =0x9 
S_MOD_TERM_FUNC_POINTERS =0xA 
S_COALESCED =0xB 
S_GB_ZEROFILL =0xC 
S_INTERPOSING =0xD 
S_16BYTE_LITERALS =0xE 
S_DTRACE_DOF =0xF 
S_LAZY_DYLIB_SYMBOL_POINTERS =0x10 
S_THREAD_LOCAL_REGULAR =0x11 
S_THREAD_LOCAL_ZEROFILL =0x12 
S_THREAD_LOCAL_VARIABLES =0x13 
S_THREAD_LOCAL_VARIABLE_POINTERS =0x14 
S_THREAD_LOCAL_INIT_FUNCTION_POINTERS =0x15 

FLAG_SECTION_TYPES ={
S_REGULAR :"S_REGULAR",
S_ZEROFILL :"S_ZEROFILL",
S_CSTRING_LITERALS :"S_CSTRING_LITERALS",
S_4BYTE_LITERALS :"S_4BYTE_LITERALS",
S_8BYTE_LITERALS :"S_8BYTE_LITERALS",
S_LITERAL_POINTERS :"S_LITERAL_POINTERS",
S_NON_LAZY_SYMBOL_POINTERS :"S_NON_LAZY_SYMBOL_POINTERS",
S_LAZY_SYMBOL_POINTERS :"S_LAZY_SYMBOL_POINTERS",
S_SYMBOL_STUBS :"S_SYMBOL_STUBS",
S_MOD_INIT_FUNC_POINTERS :"S_MOD_INIT_FUNC_POINTERS",
S_MOD_TERM_FUNC_POINTERS :"S_MOD_TERM_FUNC_POINTERS",
S_COALESCED :"S_COALESCED",
S_GB_ZEROFILL :"S_GB_ZEROFILL",
S_INTERPOSING :"S_INTERPOSING",
S_16BYTE_LITERALS :"S_16BYTE_LITERALS",
S_DTRACE_DOF :"S_DTRACE_DOF",
S_LAZY_DYLIB_SYMBOL_POINTERS :"S_LAZY_DYLIB_SYMBOL_POINTERS",
S_THREAD_LOCAL_REGULAR :"S_THREAD_LOCAL_REGULAR",
S_THREAD_LOCAL_ZEROFILL :"S_THREAD_LOCAL_ZEROFILL",
S_THREAD_LOCAL_VARIABLES :"S_THREAD_LOCAL_VARIABLES",
S_THREAD_LOCAL_VARIABLE_POINTERS :"S_THREAD_LOCAL_VARIABLE_POINTERS",
S_THREAD_LOCAL_INIT_FUNCTION_POINTERS :"S_THREAD_LOCAL_INIT_FUNCTION_POINTERS",
}

SECTION_ATTRIBUTES_USR =0xFF000000 
S_ATTR_PURE_INSTRUCTIONS =0x80000000 
S_ATTR_NO_TOC =0x40000000 
S_ATTR_STRIP_STATIC_SYMS =0x20000000 
S_ATTR_NO_DEAD_STRIP =0x10000000 
S_ATTR_LIVE_SUPPORT =0x08000000 
S_ATTR_SELF_MODIFYING_CODE =0x04000000 
S_ATTR_DEBUG =0x02000000 

SECTION_ATTRIBUTES_SYS =0x00FFFF00 
S_ATTR_SOME_INSTRUCTIONS =0x00000400 
S_ATTR_EXT_RELOC =0x00000200 
S_ATTR_LOC_RELOC =0x00000100 

FLAG_SECTION_ATTRIBUTES ={
S_ATTR_PURE_INSTRUCTIONS :"S_ATTR_PURE_INSTRUCTIONS",
S_ATTR_NO_TOC :"S_ATTR_NO_TOC",
S_ATTR_STRIP_STATIC_SYMS :"S_ATTR_STRIP_STATIC_SYMS",
S_ATTR_NO_DEAD_STRIP :"S_ATTR_NO_DEAD_STRIP",
S_ATTR_LIVE_SUPPORT :"S_ATTR_LIVE_SUPPORT",
S_ATTR_SELF_MODIFYING_CODE :"S_ATTR_SELF_MODIFYING_CODE",
S_ATTR_DEBUG :"S_ATTR_DEBUG",
S_ATTR_SOME_INSTRUCTIONS :"S_ATTR_SOME_INSTRUCTIONS",
S_ATTR_EXT_RELOC :"S_ATTR_EXT_RELOC",
S_ATTR_LOC_RELOC :"S_ATTR_LOC_RELOC",
}

SEG_PAGEZERO ="__PAGEZERO"
SEG_TEXT ="__TEXT"
SECT_TEXT ="__text"
SECT_FVMLIB_INIT0 ="__fvmlib_init0"
SECT_FVMLIB_INIT1 ="__fvmlib_init1"
SEG_DATA ="__DATA"
SECT_DATA ="__data"
SECT_BSS ="__bss"
SECT_COMMON ="__common"
SEG_OBJC ="__OBJC"
SECT_OBJC_SYMBOLS ="__symbol_table"
SECT_OBJC_MODULES ="__module_info"
SECT_OBJC_STRINGS ="__selector_strs"
SECT_OBJC_REFS ="__selector_refs"
SEG_ICON ="__ICON"
SECT_ICON_HEADER ="__header"
SECT_ICON_TIFF ="__tiff"
SEG_LINKEDIT ="__LINKEDIT"
SEG_UNIXSTACK ="__UNIXSTACK"
SEG_IMPORT ="__IMPORT"








class fvmlib (Structure ):
    _fields_ =(
    ("name",lc_str ),
    ("minor_version",mach_version_helper ),
    ("header_addr",p_uint32 ),
    )


class fvmlib_command (Structure ):
    _fields_ =fvmlib ._fields_ 

    def describe (self ):
        s ={}
        s ["header_addr"]=int (self .header_addr )
        return s 


class dylib (Structure ):
    _fields_ =(
    ("name",lc_str ),
    ("timestamp",mach_timestamp_helper ),
    ("current_version",mach_version_helper ),
    ("compatibility_version",mach_version_helper ),
    )



class dylib_command (Structure ):
    _fields_ =dylib ._fields_ 

    def describe (self ):
        s ={}
        s ["timestamp"]=str (self .timestamp )
        s ["current_version"]=str (self .current_version )
        s ["compatibility_version"]=str (self .compatibility_version )
        return s 


class sub_framework_command (Structure ):
    _fields_ =(("umbrella",lc_str ),)

    def describe (self ):
        return {}


class sub_client_command (Structure ):
    _fields_ =(("client",lc_str ),)

    def describe (self ):
        return {}


class sub_umbrella_command (Structure ):
    _fields_ =(("sub_umbrella",lc_str ),)

    def describe (self ):
        return {}


class sub_library_command (Structure ):
    _fields_ =(("sub_library",lc_str ),)

    def describe (self ):
        return {}


class prebound_dylib_command (Structure ):
    _fields_ =(("name",lc_str ),("nmodules",p_uint32 ),("linked_modules",lc_str ))

    def describe (self ):
        return {"nmodules":int (self .nmodules )}


class dylinker_command (Structure ):
    _fields_ =(("name",lc_str ),)

    def describe (self ):
        return {}


class thread_command (Structure ):
    _fields_ =(("flavor",p_uint32 ),("count",p_uint32 ))

    def describe (self ):
        s ={}
        s ["flavor"]=int (self .flavor )
        s ["count"]=int (self .count )
        return s 


class entry_point_command (Structure ):
    _fields_ =(("entryoff",p_uint64 ),("stacksize",p_uint64 ))

    def describe (self ):
        s ={}
        s ["entryoff"]=int (self .entryoff )
        s ["stacksize"]=int (self .stacksize )
        return s 


class routines_command (Structure ):
    _fields_ =(
    ("init_address",p_uint32 ),
    ("init_module",p_uint32 ),
    ("reserved1",p_uint32 ),
    ("reserved2",p_uint32 ),
    ("reserved3",p_uint32 ),
    ("reserved4",p_uint32 ),
    ("reserved5",p_uint32 ),
    ("reserved6",p_uint32 ),
    )

    def describe (self ):
        s ={}
        s ["init_address"]=int (self .init_address )
        s ["init_module"]=int (self .init_module )
        s ["reserved1"]=int (self .reserved1 )
        s ["reserved2"]=int (self .reserved2 )
        s ["reserved3"]=int (self .reserved3 )
        s ["reserved4"]=int (self .reserved4 )
        s ["reserved5"]=int (self .reserved5 )
        s ["reserved6"]=int (self .reserved6 )
        return s 


class routines_command_64 (Structure ):
    _fields_ =(
    ("init_address",p_uint64 ),
    ("init_module",p_uint64 ),
    ("reserved1",p_uint64 ),
    ("reserved2",p_uint64 ),
    ("reserved3",p_uint64 ),
    ("reserved4",p_uint64 ),
    ("reserved5",p_uint64 ),
    ("reserved6",p_uint64 ),
    )

    def describe (self ):
        s ={}
        s ["init_address"]=int (self .init_address )
        s ["init_module"]=int (self .init_module )
        s ["reserved1"]=int (self .reserved1 )
        s ["reserved2"]=int (self .reserved2 )
        s ["reserved3"]=int (self .reserved3 )
        s ["reserved4"]=int (self .reserved4 )
        s ["reserved5"]=int (self .reserved5 )
        s ["reserved6"]=int (self .reserved6 )
        return s 


class symtab_command (Structure ):
    _fields_ =(
    ("symoff",p_uint32 ),
    ("nsyms",p_uint32 ),
    ("stroff",p_uint32 ),
    ("strsize",p_uint32 ),
    )

    def describe (self ):
        s ={}
        s ["symoff"]=int (self .symoff )
        s ["nsyms"]=int (self .nsyms )
        s ["stroff"]=int (self .stroff )
        s ["strsize"]=int (self .strsize )
        return s 


class dysymtab_command (Structure ):
    _fields_ =(
    ("ilocalsym",p_uint32 ),
    ("nlocalsym",p_uint32 ),
    ("iextdefsym",p_uint32 ),
    ("nextdefsym",p_uint32 ),
    ("iundefsym",p_uint32 ),
    ("nundefsym",p_uint32 ),
    ("tocoff",p_uint32 ),
    ("ntoc",p_uint32 ),
    ("modtaboff",p_uint32 ),
    ("nmodtab",p_uint32 ),
    ("extrefsymoff",p_uint32 ),
    ("nextrefsyms",p_uint32 ),
    ("indirectsymoff",p_uint32 ),
    ("nindirectsyms",p_uint32 ),
    ("extreloff",p_uint32 ),
    ("nextrel",p_uint32 ),
    ("locreloff",p_uint32 ),
    ("nlocrel",p_uint32 ),
    )

    def describe (self ):
        dys ={}
        dys ["ilocalsym"]=int (self .ilocalsym )
        dys ["nlocalsym"]=int (self .nlocalsym )
        dys ["iextdefsym"]=int (self .iextdefsym )
        dys ["nextdefsym"]=int (self .nextdefsym )
        dys ["iundefsym"]=int (self .iundefsym )
        dys ["nundefsym"]=int (self .nundefsym )
        dys ["tocoff"]=int (self .tocoff )
        dys ["ntoc"]=int (self .ntoc )
        dys ["modtaboff"]=int (self .modtaboff )
        dys ["nmodtab"]=int (self .nmodtab )
        dys ["extrefsymoff"]=int (self .extrefsymoff )
        dys ["nextrefsyms"]=int (self .nextrefsyms )
        dys ["indirectsymoff"]=int (self .indirectsymoff )
        dys ["nindirectsyms"]=int (self .nindirectsyms )
        dys ["extreloff"]=int (self .extreloff )
        dys ["nextrel"]=int (self .nextrel )
        dys ["locreloff"]=int (self .locreloff )
        dys ["nlocrel"]=int (self .nlocrel )
        return dys 


INDIRECT_SYMBOL_LOCAL =0x80000000 
INDIRECT_SYMBOL_ABS =0x40000000 


class dylib_table_of_contents (Structure ):
    _fields_ =(("symbol_index",p_uint32 ),("module_index",p_uint32 ))


class dylib_module (Structure ):
    _fields_ =(
    ("module_name",p_uint32 ),
    ("iextdefsym",p_uint32 ),
    ("nextdefsym",p_uint32 ),
    ("irefsym",p_uint32 ),
    ("nrefsym",p_uint32 ),
    ("ilocalsym",p_uint32 ),
    ("nlocalsym",p_uint32 ),
    ("iextrel",p_uint32 ),
    ("nextrel",p_uint32 ),
    ("iinit_iterm",p_uint32 ),
    ("ninit_nterm",p_uint32 ),
    ("objc_module_info_addr",p_uint32 ),
    ("objc_module_info_size",p_uint32 ),
    )


class dylib_module_64 (Structure ):
    _fields_ =(
    ("module_name",p_uint32 ),
    ("iextdefsym",p_uint32 ),
    ("nextdefsym",p_uint32 ),
    ("irefsym",p_uint32 ),
    ("nrefsym",p_uint32 ),
    ("ilocalsym",p_uint32 ),
    ("nlocalsym",p_uint32 ),
    ("iextrel",p_uint32 ),
    ("nextrel",p_uint32 ),
    ("iinit_iterm",p_uint32 ),
    ("ninit_nterm",p_uint32 ),
    ("objc_module_info_size",p_uint32 ),
    ("objc_module_info_addr",p_uint64 ),
    )


class dylib_reference (Structure ):
    _fields_ =(
    ("isym_flags",p_uint32 ),


    )


class twolevel_hints_command (Structure ):
    _fields_ =(("offset",p_uint32 ),("nhints",p_uint32 ))

    def describe (self ):
        s ={}
        s ["offset"]=int (self .offset )
        s ["nhints"]=int (self .nhints )
        return s 


class twolevel_hint (Structure ):
    _fields_ =(
    ("isub_image_itoc",p_uint32 ),


    )


class prebind_cksum_command (Structure ):
    _fields_ =(("cksum",p_uint32 ),)

    def describe (self ):
        return {"cksum":int (self .cksum )}


class symseg_command (Structure ):
    _fields_ =(("offset",p_uint32 ),("size",p_uint32 ))

    def describe (self ):
        s ={}
        s ["offset"]=int (self .offset )
        s ["size"]=int (self .size )


class ident_command (Structure ):
    _fields_ =()

    def describe (self ):
        return {}


class fvmfile_command (Structure ):
    _fields_ =(("name",lc_str ),("header_addr",p_uint32 ))

    def describe (self ):
        return {"header_addr":int (self .header_addr )}


class uuid_command (Structure ):
    _fields_ =(("uuid",p_str16 ),)

    def describe (self ):
        return {"uuid":self .uuid .rstrip ("\x00")}


class rpath_command (Structure ):
    _fields_ =(("path",lc_str ),)

    def describe (self ):
        return {}


class linkedit_data_command (Structure ):
    _fields_ =(("dataoff",p_uint32 ),("datasize",p_uint32 ))

    def describe (self ):
        s ={}
        s ["dataoff"]=int (self .dataoff )
        s ["datasize"]=int (self .datasize )
        return s 


class version_min_command (Structure ):
    _fields_ =(
    ("version",p_uint32 ),
    ("sdk",p_uint32 ),
    )

    def describe (self ):
        v =int (self .version )
        v3 =v &0xFF 
        v =v >>8 
        v2 =v &0xFF 
        v =v >>8 
        v1 =v &0xFFFF 
        s =int (self .sdk )
        s3 =s &0xFF 
        s =s >>8 
        s2 =s &0xFF 
        s =s >>8 
        s1 =s &0xFFFF 
        return {
        "version":str (int (v1 ))+"."+str (int (v2 ))+"."+str (int (v3 )),
        "sdk":str (int (s1 ))+"."+str (int (s2 ))+"."+str (int (s3 )),
        }


class source_version_command (Structure ):
    _fields_ =(("version",p_uint64 ),)

    def describe (self ):
        v =int (self .version )
        a =v >>40 
        b =(v >>30 )&0x3FF 
        c =(v >>20 )&0x3FF 
        d =(v >>10 )&0x3FF 
        e =v &0x3FF 
        r =str (a )+"."+str (b )+"."+str (c )+"."+str (d )+"."+str (e )
        return {"version":r }


class note_command (Structure ):
    _fields_ =(("data_owner",p_str16 ),("offset",p_uint64 ),("size",p_uint64 ))


class build_version_command (Structure ):
    _fields_ =(
    ("platform",p_uint32 ),
    ("minos",p_uint32 ),
    ("sdk",p_uint32 ),
    ("ntools",p_uint32 ),
    )

    def describe (self ):
        return {}


class build_tool_version (Structure ):
    _fields_ =(("tool",p_uint32 ),("version",p_uint32 ))


class data_in_code_entry (Structure ):
    _fields_ =(("offset",p_uint32 ),("length",p_uint32 ),("kind",p_uint32 ))

    def describe (self ):
        return {"offset":self .offset ,"length":self .length ,"kind":self .kind }


DICE_KIND_DATA =0x0001 
DICE_KIND_JUMP_TABLE8 =0x0002 
DICE_KIND_JUMP_TABLE16 =0x0003 
DICE_KIND_JUMP_TABLE32 =0x0004 
DICE_KIND_ABS_JUMP_TABLE32 =0x0005 

DATA_IN_CODE_KINDS ={
DICE_KIND_DATA :"DICE_KIND_DATA",
DICE_KIND_JUMP_TABLE8 :"DICE_KIND_JUMP_TABLE8",
DICE_KIND_JUMP_TABLE16 :"DICE_KIND_JUMP_TABLE16",
DICE_KIND_JUMP_TABLE32 :"DICE_KIND_JUMP_TABLE32",
DICE_KIND_ABS_JUMP_TABLE32 :"DICE_KIND_ABS_JUMP_TABLE32",
}


class tlv_descriptor (Structure ):
    _fields_ =(
    ("thunk",p_long ),
    ("key",p_ulong ),
    ("offset",p_ulong ),
    )

    def describe (self ):
        return {"thunk":self .thunk ,"key":self .key ,"offset":self .offset }


class encryption_info_command (Structure ):
    _fields_ =(("cryptoff",p_uint32 ),("cryptsize",p_uint32 ),("cryptid",p_uint32 ))

    def describe (self ):
        s ={}
        s ["cryptoff"]=int (self .cryptoff )
        s ["cryptsize"]=int (self .cryptsize )
        s ["cryptid"]=int (self .cryptid )
        return s 


class encryption_info_command_64 (Structure ):
    _fields_ =(
    ("cryptoff",p_uint32 ),
    ("cryptsize",p_uint32 ),
    ("cryptid",p_uint32 ),
    ("pad",p_uint32 ),
    )

    def describe (self ):
        s ={}
        s ["cryptoff"]=int (self .cryptoff )
        s ["cryptsize"]=int (self .cryptsize )
        s ["cryptid"]=int (self .cryptid )
        s ["pad"]=int (self .pad )
        return s 


class dyld_info_command (Structure ):
    _fields_ =(
    ("rebase_off",p_uint32 ),
    ("rebase_size",p_uint32 ),
    ("bind_off",p_uint32 ),
    ("bind_size",p_uint32 ),
    ("weak_bind_off",p_uint32 ),
    ("weak_bind_size",p_uint32 ),
    ("lazy_bind_off",p_uint32 ),
    ("lazy_bind_size",p_uint32 ),
    ("export_off",p_uint32 ),
    ("export_size",p_uint32 ),
    )

    def describe (self ):
        dyld ={}
        dyld ["rebase_off"]=int (self .rebase_off )
        dyld ["rebase_size"]=int (self .rebase_size )
        dyld ["bind_off"]=int (self .bind_off )
        dyld ["bind_size"]=int (self .bind_size )
        dyld ["weak_bind_off"]=int (self .weak_bind_off )
        dyld ["weak_bind_size"]=int (self .weak_bind_size )
        dyld ["lazy_bind_off"]=int (self .lazy_bind_off )
        dyld ["lazy_bind_size"]=int (self .lazy_bind_size )
        dyld ["export_off"]=int (self .export_off )
        dyld ["export_size"]=int (self .export_size )
        return dyld 


class linker_option_command (Structure ):
    _fields_ =(("count",p_uint32 ),)

    def describe (self ):
        return {"count":int (self .count )}


class fileset_entry_command (Structure ):
    _fields_ =(
    ("vmaddr",p_uint64 ),
    ("fileoff",p_uint64 ),
    ("entry_id",lc_str ),
    ("reserved",p_uint32 ),
    )


LC_REGISTRY ={
LC_SEGMENT :segment_command ,
LC_IDFVMLIB :fvmlib_command ,
LC_LOADFVMLIB :fvmlib_command ,
LC_ID_DYLIB :dylib_command ,
LC_LOAD_DYLIB :dylib_command ,
LC_LOAD_WEAK_DYLIB :dylib_command ,
LC_SUB_FRAMEWORK :sub_framework_command ,
LC_SUB_CLIENT :sub_client_command ,
LC_SUB_UMBRELLA :sub_umbrella_command ,
LC_SUB_LIBRARY :sub_library_command ,
LC_PREBOUND_DYLIB :prebound_dylib_command ,
LC_ID_DYLINKER :dylinker_command ,
LC_LOAD_DYLINKER :dylinker_command ,
LC_THREAD :thread_command ,
LC_UNIXTHREAD :thread_command ,
LC_ROUTINES :routines_command ,
LC_SYMTAB :symtab_command ,
LC_DYSYMTAB :dysymtab_command ,
LC_TWOLEVEL_HINTS :twolevel_hints_command ,
LC_PREBIND_CKSUM :prebind_cksum_command ,
LC_SYMSEG :symseg_command ,
LC_IDENT :ident_command ,
LC_FVMFILE :fvmfile_command ,
LC_SEGMENT_64 :segment_command_64 ,
LC_ROUTINES_64 :routines_command_64 ,
LC_UUID :uuid_command ,
LC_RPATH :rpath_command ,
LC_CODE_SIGNATURE :linkedit_data_command ,
LC_CODE_SEGMENT_SPLIT_INFO :linkedit_data_command ,
LC_REEXPORT_DYLIB :dylib_command ,
LC_LAZY_LOAD_DYLIB :dylib_command ,
LC_ENCRYPTION_INFO :encryption_info_command ,
LC_DYLD_INFO :dyld_info_command ,
LC_DYLD_INFO_ONLY :dyld_info_command ,
LC_LOAD_UPWARD_DYLIB :dylib_command ,
LC_VERSION_MIN_MACOSX :version_min_command ,
LC_VERSION_MIN_IPHONEOS :version_min_command ,
LC_FUNCTION_STARTS :linkedit_data_command ,
LC_DYLD_ENVIRONMENT :dylinker_command ,
LC_MAIN :entry_point_command ,
LC_DATA_IN_CODE :linkedit_data_command ,
LC_SOURCE_VERSION :source_version_command ,
LC_DYLIB_CODE_SIGN_DRS :linkedit_data_command ,
LC_ENCRYPTION_INFO_64 :encryption_info_command_64 ,
LC_LINKER_OPTION :linker_option_command ,
LC_LINKER_OPTIMIZATION_HINT :linkedit_data_command ,
LC_VERSION_MIN_TVOS :version_min_command ,
LC_VERSION_MIN_WATCHOS :version_min_command ,
LC_NOTE :note_command ,
LC_BUILD_VERSION :build_version_command ,
LC_DYLD_EXPORTS_TRIE :linkedit_data_command ,
LC_DYLD_CHAINED_FIXUPS :linkedit_data_command ,
LC_FILESET_ENTRY :fileset_entry_command ,
}

LC_NAMES ={
LC_SEGMENT :"LC_SEGMENT",
LC_IDFVMLIB :"LC_IDFVMLIB",
LC_LOADFVMLIB :"LC_LOADFVMLIB",
LC_ID_DYLIB :"LC_ID_DYLIB",
LC_LOAD_DYLIB :"LC_LOAD_DYLIB",
LC_LOAD_WEAK_DYLIB :"LC_LOAD_WEAK_DYLIB",
LC_SUB_FRAMEWORK :"LC_SUB_FRAMEWORK",
LC_SUB_CLIENT :"LC_SUB_CLIENT",
LC_SUB_UMBRELLA :"LC_SUB_UMBRELLA",
LC_SUB_LIBRARY :"LC_SUB_LIBRARY",
LC_PREBOUND_DYLIB :"LC_PREBOUND_DYLIB",
LC_ID_DYLINKER :"LC_ID_DYLINKER",
LC_LOAD_DYLINKER :"LC_LOAD_DYLINKER",
LC_THREAD :"LC_THREAD",
LC_UNIXTHREAD :"LC_UNIXTHREAD",
LC_ROUTINES :"LC_ROUTINES",
LC_SYMTAB :"LC_SYMTAB",
LC_DYSYMTAB :"LC_DYSYMTAB",
LC_TWOLEVEL_HINTS :"LC_TWOLEVEL_HINTS",
LC_PREBIND_CKSUM :"LC_PREBIND_CKSUM",
LC_SYMSEG :"LC_SYMSEG",
LC_IDENT :"LC_IDENT",
LC_FVMFILE :"LC_FVMFILE",
LC_SEGMENT_64 :"LC_SEGMENT_64",
LC_ROUTINES_64 :"LC_ROUTINES_64",
LC_UUID :"LC_UUID",
LC_RPATH :"LC_RPATH",
LC_CODE_SIGNATURE :"LC_CODE_SIGNATURE",
LC_CODE_SEGMENT_SPLIT_INFO :"LC_CODE_SEGMENT_SPLIT_INFO",
LC_REEXPORT_DYLIB :"LC_REEXPORT_DYLIB",
LC_LAZY_LOAD_DYLIB :"LC_LAZY_LOAD_DYLIB",
LC_ENCRYPTION_INFO :"LC_ENCRYPTION_INFO",
LC_DYLD_INFO :"LC_DYLD_INFO",
LC_DYLD_INFO_ONLY :"LC_DYLD_INFO_ONLY",
LC_LOAD_UPWARD_DYLIB :"LC_LOAD_UPWARD_DYLIB",
LC_VERSION_MIN_MACOSX :"LC_VERSION_MIN_MACOSX",
LC_VERSION_MIN_IPHONEOS :"LC_VERSION_MIN_IPHONEOS",
LC_FUNCTION_STARTS :"LC_FUNCTION_STARTS",
LC_DYLD_ENVIRONMENT :"LC_DYLD_ENVIRONMENT",
LC_MAIN :"LC_MAIN",
LC_DATA_IN_CODE :"LC_DATA_IN_CODE",
LC_SOURCE_VERSION :"LC_SOURCE_VERSION",
LC_DYLIB_CODE_SIGN_DRS :"LC_DYLIB_CODE_SIGN_DRS",
LC_LINKER_OPTIMIZATION_HINT :"LC_LINKER_OPTIMIZATION_HINT",
LC_VERSION_MIN_TVOS :"LC_VERSION_MIN_TVOS",
LC_VERSION_MIN_WATCHOS :"LC_VERSION_MIN_WATCHOS",
LC_NOTE :"LC_NOTE",
LC_BUILD_VERSION :"LC_BUILD_VERSION",
LC_DYLD_EXPORTS_TRIE :"LC_DYLD_EXPORTS_TRIE",
LC_DYLD_CHAINED_FIXUPS :"LC_DYLD_CHAINED_FIXUPS",
LC_ENCRYPTION_INFO_64 :"LC_ENCRYPTION_INFO_64",
LC_LINKER_OPTION :"LC_LINKER_OPTION",
LC_PREPAGE :"LC_PREPAGE",
LC_FILESET_ENTRY :"LC_FILESET_ENTRY",
}



class n_un (p_int32 ):
    pass 


class nlist (Structure ):
    _fields_ =(
    ("n_un",n_un ),
    ("n_type",p_uint8 ),
    ("n_sect",p_uint8 ),
    ("n_desc",p_short ),
    ("n_value",p_uint32 ),
    )


class nlist_64 (Structure ):
    _fields_ =[
    ("n_un",n_un ),
    ("n_type",p_uint8 ),
    ("n_sect",p_uint8 ),
    ("n_desc",p_short ),
    ("n_value",p_int64 ),
    ]


N_STAB =0xE0 
N_PEXT =0x10 
N_TYPE =0x0E 
N_EXT =0x01 

N_UNDF =0x0 
N_ABS =0x2 
N_SECT =0xE 
N_PBUD =0xC 
N_INDR =0xA 

NO_SECT =0 
MAX_SECT =255 


class relocation_info (Structure ):
    _fields_ =(("r_address",p_uint32 ),("_r_bitfield",p_uint32 ))

    def _describe (self ):
        return (("r_address",self .r_address ),("_r_bitfield",self ._r_bitfield ))


def GET_COMM_ALIGN (n_desc ):
    return (n_desc >>8 )&0x0F 


def SET_COMM_ALIGN (n_desc ,align ):
    return (n_desc &0xF0FF )|((align &0x0F )<<8 )


REFERENCE_TYPE =0xF 
REFERENCE_FLAG_UNDEFINED_NON_LAZY =0 
REFERENCE_FLAG_UNDEFINED_LAZY =1 
REFERENCE_FLAG_DEFINED =2 
REFERENCE_FLAG_PRIVATE_DEFINED =3 
REFERENCE_FLAG_PRIVATE_UNDEFINED_NON_LAZY =4 
REFERENCE_FLAG_PRIVATE_UNDEFINED_LAZY =5 

REFERENCED_DYNAMICALLY =0x0010 


def GET_LIBRARY_ORDINAL (n_desc ):
    return ((n_desc )>>8 )&0xFF 


def SET_LIBRARY_ORDINAL (n_desc ,ordinal ):
    return ((n_desc )&0x00FF )|(((ordinal &0xFF )<<8 ))


SELF_LIBRARY_ORDINAL =0x0 
MAX_LIBRARY_ORDINAL =0xFD 
DYNAMIC_LOOKUP_ORDINAL =0xFE 
EXECUTABLE_ORDINAL =0xFF 

N_NO_DEAD_STRIP =0x0020 
N_DESC_DISCARDED =0x0020 
N_WEAK_REF =0x0040 
N_WEAK_DEF =0x0080 
N_REF_TO_WEAK =0x0080 
N_ARM_THUMB_DEF =0x0008 
N_SYMBOL_RESOLVER =0x0100 
N_ALT_ENTRY =0x0200 


FAT_MAGIC =0xCAFEBABE 
FAT_CIGAM =0xBEBAFECA 
FAT_MAGIC_64 =0xCAFEBABF 
FAT_CIGAM_64 =0xBFBAFECA 


class fat_header (Structure ):
    _fields_ =(("magic",p_uint32 ),("nfat_arch",p_uint32 ))


class fat_arch (Structure ):
    _fields_ =(
    ("cputype",cpu_type_t ),
    ("cpusubtype",cpu_subtype_t ),
    ("offset",p_uint32 ),
    ("size",p_uint32 ),
    ("align",p_uint32 ),
    )


class fat_arch64 (Structure ):
    _fields_ =(
    ("cputype",cpu_type_t ),
    ("cpusubtype",cpu_subtype_t ),
    ("offset",p_uint64 ),
    ("size",p_uint64 ),
    ("align",p_uint32 ),
    ("reserved",p_uint32 ),
    )


REBASE_TYPE_POINTER =1 
REBASE_TYPE_TEXT_ABSOLUTE32 =2 
REBASE_TYPE_TEXT_PCREL32 =3 

REBASE_OPCODE_MASK =0xF0 
REBASE_IMMEDIATE_MASK =0x0F 
REBASE_OPCODE_DONE =0x00 
REBASE_OPCODE_SET_TYPE_IMM =0x10 
REBASE_OPCODE_SET_SEGMENT_AND_OFFSET_ULEB =0x20 
REBASE_OPCODE_ADD_ADDR_ULEB =0x30 
REBASE_OPCODE_ADD_ADDR_IMM_SCALED =0x40 
REBASE_OPCODE_DO_REBASE_IMM_TIMES =0x50 
REBASE_OPCODE_DO_REBASE_ULEB_TIMES =0x60 
REBASE_OPCODE_DO_REBASE_ADD_ADDR_ULEB =0x70 
REBASE_OPCODE_DO_REBASE_ULEB_TIMES_SKIPPING_ULEB =0x80 

BIND_TYPE_POINTER =1 
BIND_TYPE_TEXT_ABSOLUTE32 =2 
BIND_TYPE_TEXT_PCREL32 =3 

BIND_SPECIAL_DYLIB_SELF =0 
BIND_SPECIAL_DYLIB_MAIN_EXECUTABLE =-1 
BIND_SPECIAL_DYLIB_FLAT_LOOKUP =-2 

BIND_SYMBOL_FLAGS_WEAK_IMPORT =0x1 
BIND_SYMBOL_FLAGS_NON_WEAK_DEFINITION =0x8 

BIND_OPCODE_MASK =0xF0 
BIND_IMMEDIATE_MASK =0x0F 
BIND_OPCODE_DONE =0x00 
BIND_OPCODE_SET_DYLIB_ORDINAL_IMM =0x10 
BIND_OPCODE_SET_DYLIB_ORDINAL_ULEB =0x20 
BIND_OPCODE_SET_DYLIB_SPECIAL_IMM =0x30 
BIND_OPCODE_SET_SYMBOL_TRAILING_FLAGS_IMM =0x40 
BIND_OPCODE_SET_TYPE_IMM =0x50 
BIND_OPCODE_SET_ADDEND_SLEB =0x60 
BIND_OPCODE_SET_SEGMENT_AND_OFFSET_ULEB =0x70 
BIND_OPCODE_ADD_ADDR_ULEB =0x80 
BIND_OPCODE_DO_BIND =0x90 
BIND_OPCODE_DO_BIND_ADD_ADDR_ULEB =0xA0 
BIND_OPCODE_DO_BIND_ADD_ADDR_IMM_SCALED =0xB0 
BIND_OPCODE_DO_BIND_ULEB_TIMES_SKIPPING_ULEB =0xC0 

EXPORT_SYMBOL_FLAGS_KIND_MASK =0x03 
EXPORT_SYMBOL_FLAGS_KIND_REGULAR =0x00 
EXPORT_SYMBOL_FLAGS_KIND_THREAD_LOCAL =0x01 
EXPORT_SYMBOL_FLAGS_WEAK_DEFINITION =0x04 
EXPORT_SYMBOL_FLAGS_REEXPORT =0x08 
EXPORT_SYMBOL_FLAGS_STUB_AND_RESOLVER =0x10 

PLATFORM_MACOS =1 
PLATFORM_IOS =2 
PLATFORM_TVOS =3 
PLATFORM_WATCHOS =4 
PLATFORM_BRIDGEOS =5 
PLATFORM_IOSMAC =6 
PLATFORM_MACCATALYST =6 
PLATFORM_IOSSIMULATOR =7 
PLATFORM_TVOSSIMULATOR =8 
PLATFORM_WATCHOSSIMULATOR =9 

PLATFORM_NAMES ={
PLATFORM_MACOS :"macOS",
PLATFORM_IOS :"iOS",
PLATFORM_TVOS :"tvOS",
PLATFORM_WATCHOS :"watchOS",
PLATFORM_BRIDGEOS :"bridgeOS",
PLATFORM_MACCATALYST :"catalyst",
PLATFORM_IOSSIMULATOR :"iOS simulator",
PLATFORM_TVOSSIMULATOR :"tvOS simulator",
PLATFORM_WATCHOSSIMULATOR :"watchOS simulator",
}

TOOL_CLANG =1 
TOOL_SWIFT =2 
TOOL_LD =3 

TOOL_NAMES ={TOOL_CLANG :"clang",TOOL_SWIFT :"swift",TOOL_LD :"ld"}



"""
Utilities for creating dot output from a MachOGraph
"""

from collections import deque 

try :
    from itertools import imap 
except ImportError :
    imap =map 

__all__ =["itergraphreport"]


def itergraphreport (nodes ,describe_edge ,name ="G"):
    edges =deque ()
    nodetoident ={}

    def nodevisitor (node ,data ,outgoing ,incoming ):
        return {"label":str (node )}

    def edgevisitor (edge ,data ,head ,tail ):
        return {}

    yield "digraph %s {\n"%(name ,)
    attr ={"rankdir":"LR","concentrate":"true"}
    cpatt ='%s="%s"'
    for item in attr .items ():
        yield "\t%s;\n"%(cpatt %item ,)


    for node ,data ,_outgoing ,_incoming in nodes :
        nodetoident [node ]=getattr (data ,"identifier",node )


    for node ,data ,outgoing ,incoming in nodes :

        for edge in imap (describe_edge ,outgoing ):
            edges .append (edge )


        yield '\t"%s" [%s];\n'%(
        node ,
        ",".join (
        [
        (cpatt %item )
        for item in nodevisitor (node ,data ,outgoing ,incoming ).items ()
        ]
        ),
        )

    graph =[]

    while edges :
        edge ,data ,head ,tail =edges .popleft ()
        if data in ("run_file","load_dylib"):
            graph .append ((edge ,data ,head ,tail ))

    def do_graph (edges ,tabs ):
        edgestr =tabs +'"%s" -> "%s" [%s];\n'

        for edge ,data ,head ,tail in edges :
            attribs =edgevisitor (edge ,data ,head ,tail )
            yield edgestr %(
            head ,
            tail ,
            ",".join ([(cpatt %item )for item in attribs .items ()]),
            )

    for s in do_graph (graph ,"\t"):
        yield s 

    yield "}\n"



"""
virtualenv installs a wrapper for the real distutils into the
virtual environment. Ignore that wrapper, but go for the real
distutils instead

This recipe is rather compilicated and definitely not a
good model for other recipes!!!
"""
from __future__ import absolute_import 

try :
    import imp 
except ImportError :
    from modulegraph import _imp as imp 

import os 
import sys 

from modulegraph .modulegraph import (
CompiledModule ,
MissingModule ,
Package ,
SourceModule ,
find_module ,
)


def retry_import (mf ,m ):
    """
    Try to reimport 'm', which should be a MissingModule
    """
    if "."in m .identifier :
        pname ,partname =m .identifier .rsplit (".",1 )
        parent =mf .findNode (pname )
    else :
        parent =None 
        partname =m .identifier 



    def fmod (name ,path ,parent ):
        if path is None :
            if name in sys .builtin_module_names :
                return (None ,None ,("","",imp .C_BUILTIN ))

            path =mf .path 

        fp ,buf ,stuff =find_module (name ,path )
        if buf :
            buf =os .path .realpath (buf )
        return (fp ,buf ,stuff )

    try :
        fp ,pathname ,stuff =fmod (partname ,parent and parent .packagepath ,parent )
    except ImportError :
        return 

    if stuff [-1 ]==imp .PKG_DIRECTORY :
        m .__class__ =Package 
    elif stuff [-1 ]==imp .PY_SOURCE :
        m .__class__ =SourceModule 
    else :
        m .__class__ =CompiledModule 



    if hasattr (mf ,"load_module"):
        m =mf .load_module (m .identifier ,fp ,pathname ,stuff )
    else :
        m =mf ._load_module (m .identifier ,fp ,pathname ,stuff )

    if parent :
        mf .createReference (m ,parent )
        parent [partname ]=m 
    return m 


def check (cmd ,mf ):
    m =mf .findNode ("distutils")
    if m is None or m .filename is None :
        return None 

    with open (m .filename ,"r")as fp :
        contents =fp .read ()
    if "virtualenv"in contents :

        mos =mf .findNode ("os")
        if mos is None or mos .filename is None :
            raise ValueError ("Where is the os module")

        m .filename =os .path .join (
        os .path .dirname (mos .filename ),"distutils","__init__.py"
        )
        with open (m .filename )as fp :
            source =fp .read ()+"\n"
        m .code =co =compile (source ,m .filename ,"exec")
        m .packagepath =[os .path .dirname (m .filename )]

        if mf .replace_paths :
            co =mf .replace_paths_in_code (co )



        if hasattr (mf ,"scan_code"):
            mf .scan_code (co ,m )
        else :
            mf ._scan_code (co ,m )




        for m in mf .flatten ():
            if isinstance (m ,MissingModule ):
                if m .identifier .startswith ("distutils."):



                    retry_import (mf ,m )

    return {}



import sys 

SIX_TAB ={

"configparser":("ConfigParser","configparser"),
"copyreg":("copy_reg","copyreg"),
"cPickle":("cPickle","pickle"),
"cStringIO":("cStringIO","io"),
"dbm_gnu":("gdbm","dbm.gnu"),
"_dummy_thread":("dummy_thread","_dummy_thread"),
"email_mime_multipart":("email.MIMEMultipart","email.mime.multipart"),
"email_mime_nonmultipart":("email.MIMENonMultipart","email.mime.nonmultipart"),
"email_mime_text":("email.MIMEText","email.mime.text"),
"email_mime_base":("email.MIMEBase","email.mime.base"),
"filter":("itertools",None ),
"filterfalse":("itertools","itertools"),
"getcwd":("os","os"),
"getcwdb":("os","os"),
"http_cookiejar":("cookielib","http.cookiejar"),
"http_cookies":("Cookie","http.cookies"),
"html_entities":("htmlentitydefs","html.entities"),
"html_parser":("HTMLParser","html.parser"),
"http_client":("httplib","http.client"),
"BaseHTTPServer":("BaseHTTPServer","http.server"),
"CGIHTTPServer":("CGIHTTPServer","http.server"),
"SimpleHTTPServer":("SimpleHTTPServer","http.server"),
"intern":(None ,"sys"),
"map":("itertools",None ),
"queue":("Queue","queue"),
"reduce":(None ,"functools"),
"reload_module":(None ,"importlib"),
"reprlib":("repr","reprlib"),
"shlex_quote":("pipes","shlex"),
"socketserver":("SocketServer","socketserver"),
"_thread":("thread","_thread"),
"tkinter":("Tkinter","tkinter"),
"tkinter_dialog":("Dialog","tkinter.dialog"),
"tkinter_filedialog":("FileDialog","tkinter.FileDialog"),
"tkinter_scrolledtext":("ScrolledText","tkinter.scrolledtext"),
"tkinter_simpledialog":("SimpleDialog","tkinter.simpledialog"),
"tkinter_ttk":("ttk","tkinter.ttk"),
"tkinter_tix":("Tix","tkinter.tix"),
"tkinter_constants":("Tkconstants","tkinter.constants"),
"tkinter_dnd":("Tkdnd","tkinter.dnd"),
"tkinter_colorchooser":("tkColorChooser","tkinter.colorchooser"),
"tkinter_commondialog":("tkCommonDialog","tkinter.commondialog"),
"tkinter_tkfiledialog":("tkFileDialog","tkinter.filedialog"),
"tkinter_font":("tkFont","tkinter.font"),
"tkinter_messagebox":("tkMessageBox","tkinter.messagebox"),
"tkinter_tksimpledialog":("tkSimpleDialog","tkinter.simpledialog"),
"urllib.robotparser":("robotparser","urllib.robotparser"),
"urllib_robotparser":("robotparser","urllib.robotparser"),
"UserDict":("UserDict","collections"),
"UserList":("UserList","collections"),
"UserString":("UserString","collections"),
"winreg":("_winreg","winreg"),
"xmlrpc_client":("xmlrpclib","xmlrpc.client"),
"xmlrpc_server":("SimpleXMLRPCServer","xmlrpc.server"),
"zip":("itertools",None ),
"zip_longest":("itertools","itertools"),
"urllib.parse":(("urlparse","urllib"),"urllib.parse"),
"urllib.error":(("urllib","urllib2"),"urllib.error"),
"urllib.request":(("urllib","urllib2"),"urllib.request"),
"urllib.response":("urllib","urllib.request"),
}


def check (cmd ,mf ):
    found =False 

    six_moves =["six.moves"]



    for nm in mf .graph .node_list ():
        if not isinstance (nm ,str ):
            continue 
        if nm .endswith (".six.moves"):
            six_moves .append (nm )

    for mod in six_moves :
        m =mf .findNode (mod )
        if m is None :
            continue 











        for submod in SIX_TAB :
            if submod .startswith ("tkinter"):


                continue 

            if sys .version_info [0 ]==2 :
                alt =SIX_TAB [submod ][0 ]
            else :
                alt =SIX_TAB [submod ][1 ]

            if alt is None :
                continue 

            elif not isinstance (alt ,tuple ):
                alt =(alt ,)

            for nm in alt :
                try :
                    mf .import_hook (nm ,m )
                    found =True 
                except ImportError :
                    pass 



        for submod in SIX_TAB :
            if not submod .startswith ("tkinter"):
                continue 

            name =mod +"."+submod 
            m =mf .findNode (name )
            if m is not None :
                if sys .version_info [0 ]==2 :
                    alt =SIX_TAB [submod ][0 ]
                else :
                    alt =SIX_TAB [submod ][1 ]

                if alt is None :
                    continue 

                elif not isinstance (alt ,tuple ):
                    alt =(alt ,)

                for nm in alt :
                    mf .import_hook (nm ,m )
                    found =True 

    if found :
        return {}

    else :
        return None 



"""
sys.argv emulation

This module starts a basic event loop to collect file- and url-open
AppleEvents. Those get converted to strings and stuffed into sys.argv.
When that is done we continue starting the application.

This is a workaround to convert scripts that expect filenames on the
command-line to work in a GUI environment. GUI applications should not
use this feature.

NOTE: This module uses ctypes and not the Carbon modules in the stdlib
because the latter don't work in 64-bit mode and are also not available
with python 3.x.
"""

import ctypes 
import struct 
import sys 
import time 


class AEDesc (ctypes .Structure ):
    _fields_ =[
    ("descKey",ctypes .c_int ),
    ("descContent",ctypes .c_void_p ),
    ]


class EventTypeSpec (ctypes .Structure ):
    _fields_ =[
    ("eventClass",ctypes .c_int ),
    ("eventKind",ctypes .c_uint ),
    ]


def _ctypes_setup ():
    carbon =ctypes .CDLL ("/System/Library/Carbon.framework/Carbon")




    ae_callback =ctypes .CFUNCTYPE (
    ctypes .c_int ,ctypes .c_void_p ,ctypes .c_void_p ,ctypes .c_void_p 
    )
    carbon .AEInstallEventHandler .argtypes =[
    ctypes .c_int ,
    ctypes .c_int ,
    ae_callback ,
    ctypes .c_void_p ,
    ctypes .c_char ,
    ]
    carbon .AERemoveEventHandler .argtypes =[
    ctypes .c_int ,
    ctypes .c_int ,
    ae_callback ,
    ctypes .c_char ,
    ]

    carbon .AEProcessEvent .restype =ctypes .c_int 
    carbon .AEProcessEvent .argtypes =[ctypes .c_void_p ]

    carbon .ReceiveNextEvent .restype =ctypes .c_int 
    carbon .ReceiveNextEvent .argtypes =[
    ctypes .c_long ,
    ctypes .POINTER (EventTypeSpec ),
    ctypes .c_double ,
    ctypes .c_char ,
    ctypes .POINTER (ctypes .c_void_p ),
    ]

    carbon .AEGetParamDesc .restype =ctypes .c_int 
    carbon .AEGetParamDesc .argtypes =[
    ctypes .c_void_p ,
    ctypes .c_int ,
    ctypes .c_int ,
    ctypes .POINTER (AEDesc ),
    ]

    carbon .AECountItems .restype =ctypes .c_int 
    carbon .AECountItems .argtypes =[
    ctypes .POINTER (AEDesc ),
    ctypes .POINTER (ctypes .c_long ),
    ]

    carbon .AEGetNthDesc .restype =ctypes .c_int 
    carbon .AEGetNthDesc .argtypes =[
    ctypes .c_void_p ,
    ctypes .c_long ,
    ctypes .c_int ,
    ctypes .c_void_p ,
    ctypes .c_void_p ,
    ]

    carbon .AEGetDescDataSize .restype =ctypes .c_int 
    carbon .AEGetDescDataSize .argtypes =[ctypes .POINTER (AEDesc )]

    carbon .AEGetDescData .restype =ctypes .c_int 
    carbon .AEGetDescData .argtypes =[
    ctypes .POINTER (AEDesc ),
    ctypes .c_void_p ,
    ctypes .c_int ,
    ]

    carbon .FSRefMakePath .restype =ctypes .c_int 
    carbon .FSRefMakePath .argtypes =[ctypes .c_void_p ,ctypes .c_void_p ,ctypes .c_uint ]

    return carbon 


def _run_argvemulator (timeout =60 ):


    carbon =_ctypes_setup ()


    running =[True ]

    timeout =[timeout ]


    ae_callback =carbon .AEInstallEventHandler .argtypes [2 ]

    (kAEInternetSuite ,)=struct .unpack (">i",b"GURL")
    (kAEISGetURL ,)=struct .unpack (">i",b"GURL")
    (kCoreEventClass ,)=struct .unpack (">i",b"aevt")
    (kAEOpenApplication ,)=struct .unpack (">i",b"oapp")
    (kAEOpenDocuments ,)=struct .unpack (">i",b"odoc")
    (keyDirectObject ,)=struct .unpack (">i",b"----")
    (typeAEList ,)=struct .unpack (">i",b"list")
    (typeChar ,)=struct .unpack (">i",b"TEXT")
    (typeFSRef ,)=struct .unpack (">i",b"fsrf")
    FALSE =b"\0"
    TRUE =b"\1"
    eventLoopTimedOutErr =-9875 

    (kEventClassAppleEvent ,)=struct .unpack (">i",b"eppc")
    kEventAppleEvent =1 

    @ae_callback 
    def open_app_handler (message ,reply ,refcon ):







        timeout [0 ]=min (timeout [0 ],time .time ()-start +2 )
        return 0 

    carbon .AEInstallEventHandler (
    kCoreEventClass ,kAEOpenApplication ,open_app_handler ,0 ,FALSE 
    )

    @ae_callback 
    def open_file_handler (message ,reply ,refcon ):
        listdesc =AEDesc ()
        sts =carbon .AEGetParamDesc (
        message ,keyDirectObject ,typeAEList ,ctypes .byref (listdesc )
        )
        if sts !=0 :
            print ("argvemulator warning: cannot unpack open document event")
            running [0 ]=False 
            return 

        item_count =ctypes .c_long ()
        sts =carbon .AECountItems (ctypes .byref (listdesc ),ctypes .byref (item_count ))
        if sts !=0 :
            print ("argvemulator warning: cannot unpack open document event")
            running [0 ]=False 
            return 

        desc =AEDesc ()
        for i in range (item_count .value ):
            sts =carbon .AEGetNthDesc (
            ctypes .byref (listdesc ),i +1 ,typeFSRef ,0 ,ctypes .byref (desc )
            )
            if sts !=0 :
                print ("argvemulator warning: cannot unpack open document event")
                running [0 ]=False 
                return 

            sz =carbon .AEGetDescDataSize (ctypes .byref (desc ))
            buf =ctypes .create_string_buffer (sz )
            sts =carbon .AEGetDescData (ctypes .byref (desc ),buf ,sz )
            if sts !=0 :
                print ("argvemulator warning: cannot extract open document event")
                continue 

            fsref =buf 

            buf =ctypes .create_string_buffer (1024 )
            sts =carbon .FSRefMakePath (ctypes .byref (fsref ),buf ,1023 )
            if sts !=0 :
                print ("argvemulator warning: cannot extract open document event")
                continue 

            if sys .version_info [0 ]>2 :
                sys .argv .append (buf .value .decode ("utf-8"))
            else :
                sys .argv .append (buf .value )

        running [0 ]=False 
        return 0 

    carbon .AEInstallEventHandler (
    kCoreEventClass ,kAEOpenDocuments ,open_file_handler ,0 ,FALSE 
    )

    @ae_callback 
    def open_url_handler (message ,reply ,refcon ):
        listdesc =AEDesc ()
        ok =carbon .AEGetParamDesc (
        message ,keyDirectObject ,typeAEList ,ctypes .byref (listdesc )
        )
        if ok !=0 :
            print ("argvemulator warning: cannot unpack open document event")
            running [0 ]=False 
            return 

        item_count =ctypes .c_long ()
        sts =carbon .AECountItems (ctypes .byref (listdesc ),ctypes .byref (item_count ))
        if sts !=0 :
            print ("argvemulator warning: cannot unpack open url event")
            running [0 ]=False 
            return 

        desc =AEDesc ()
        for i in range (item_count .value ):
            sts =carbon .AEGetNthDesc (
            ctypes .byref (listdesc ),i +1 ,typeChar ,0 ,ctypes .byref (desc )
            )
            if sts !=0 :
                print ("argvemulator warning: cannot unpack open URL event")
                running [0 ]=False 
                return 

            sz =carbon .AEGetDescDataSize (ctypes .byref (desc ))
            buf =ctypes .create_string_buffer (sz )
            sts =carbon .AEGetDescData (ctypes .byref (desc ),buf ,sz )
            if sts !=0 :
                print ("argvemulator warning: cannot extract open URL event")

            else :
                if sys .version_info [0 ]>2 :
                    sys .argv .append (buf .value .decode ("utf-8"))
                else :
                    sys .argv .append (buf .value )

        running [0 ]=False 
        return 0 

    carbon .AEInstallEventHandler (
    kAEInternetSuite ,kAEISGetURL ,open_url_handler ,0 ,FALSE 
    )


    if len (sys .argv )>1 and sys .argv [1 ].startswith ("-psn_"):
        del sys .argv [1 ]

    start =time .time ()
    now =time .time ()
    eventType =EventTypeSpec ()
    eventType .eventClass =kEventClassAppleEvent 
    eventType .eventKind =kEventAppleEvent 

    while running [0 ]and now -start <timeout [0 ]:
        event =ctypes .c_void_p ()

        sts =carbon .ReceiveNextEvent (
        1 ,
        ctypes .byref (eventType ),
        start +timeout [0 ]-now ,
        TRUE ,
        ctypes .byref (event ),
        )

        if sts ==eventLoopTimedOutErr :
            break 

        elif sts !=0 :
            print ("argvemulator warning: fetching events failed")
            break 

        sts =carbon .AEProcessEvent (event )
        if sts !=0 :
            print ("argvemulator warning: processing events failed")
            break 

    carbon .AERemoveEventHandler (
    kCoreEventClass ,kAEOpenApplication ,open_app_handler ,FALSE 
    )
    carbon .AERemoveEventHandler (
    kCoreEventClass ,kAEOpenDocuments ,open_file_handler ,FALSE 
    )
    carbon .AERemoveEventHandler (kAEInternetSuite ,kAEISGetURL ,open_url_handler ,FALSE )


def _argv_emulation ():
    import os 


    if os .environ .get ("_PY2APP_LAUNCHED_"):
        _run_argvemulator ()


_argv_emulation ()



import sys 
import os 
import numpy as np 
from PySide6 .QtWidgets import (QApplication ,QMainWindow ,QWidget ,QVBoxLayout ,QHBoxLayout ,QPushButton ,QListWidget ,QSpinBox ,QDoubleSpinBox ,QLabel ,QCheckBox ,QFileDialog ,QTabWidget ,QSizePolicy ,QGridLayout ,QSpinBox ,QMessageBox ,QLineEdit ,QFormLayout ,QHBoxLayout ,QStyleFactory ,QComboBox ,QScrollArea )
from PySide6 .QtCore import Qt 
from matplotlib .backends .backend_qt5agg import FigureCanvasQTAgg as FigureCanvas 
from matplotlib .figure import Figure 
from matplotlib .backends .backend_qt5agg import NavigationToolbar2QT as NavigationToolbar 
from functools import partial 
import seaborn as sns 
import matplotlib .pyplot as plt 
from matplotlib .ticker import ScalarFormatter 
from PySide6 .QtGui import QPalette ,QColor ,QFont 
import math 

class MplCanvas (FigureCanvas ):
    def __init__ (self ,width =5 ,height =4 ,dpi =300 ):
        fig =Figure (figsize =(width ,height ),dpi =dpi )
        self .axes =fig .add_subplot (111 )
        super ().__init__ (fig )

class PlottingApplication (QMainWindow ):
    def __init__ (self ):
        super ().__init__ ()
        self .setWindowTitle ("SD Frequency and multi-plots")
        self .histogram_bins =100 
        self .scatter_marker_size =3 
        self .global_limits ={}
        self .filtered_data ={}
        self .initUI ()

    def initUI (self ):

        mainWidget =QWidget ()
        self .setCentralWidget (mainWidget )
        mainLayout =QHBoxLayout (mainWidget )


        self .leftPanel =QTabWidget ()
        fileSelectionTab =QWidget ()
        self .plotOptionsTab =QWidget ()
        databaseTab =QWidget ()

        self .leftPanel .addTab (fileSelectionTab ,"Files")
        self .leftPanel .addTab (self .plotOptionsTab ,"Options")
        self .leftPanel .addTab (databaseTab ,"Database")


        self .rightPanel =QTabWidget ()

        mainLayout .addWidget (self .leftPanel ,25 )
        mainLayout .addWidget (self .rightPanel ,75 )

        self .setupFileSelectionTab (fileSelectionTab )
        self .setupPlotOptionsTab (self .plotOptionsTab )
        self .setupDatabaseTab (databaseTab )



    def calculateGridSize (self ,numberOfPlots ):


        gridSize =math .ceil (math .sqrt (numberOfPlots ))




        nrows =gridSize 
        ncols =gridSize 


        if gridSize *(gridSize -1 )>=numberOfPlots :
            ncols -=1 

        return nrows ,ncols 



    def setupFileSelectionTab (self ,tab ):
        layout =QVBoxLayout (tab )

        headerLabel =QLabel ("SD Frequency and multi-plots\nshankar.dutt@anu.edu.au")
        headerLabel .setWordWrap (True )
        headerLabel .setAlignment (Qt .AlignmentFlag .AlignCenter )
        headerLabel .setFont (QFont ('Arial',18 ,QFont .Weight .Bold ))
        headerLabel .setStyleSheet ("QLabel { color : white; margin-top: 20px; margin-bottom: 20px; }")


        app .setStyleSheet ("""
        QPushButton {
            background-color: #2b5b84;
            color: white;
            border-radius: 5px;
            padding: 10px;
        }
        QPushButton:hover {
            background-color: #1d4560;
        }
        """)

        layout .addWidget (headerLabel )

        selectFolderBtn =QPushButton ("Select Folder")
        selectFolderBtn .clicked .connect (self .selectFolder )
        layout .addWidget (selectFolderBtn )

        self .fileListWidget =QListWidget ()
        self .fileListWidget .setSelectionMode (QListWidget .SelectionMode .MultiSelection )
        layout .addWidget (self .fileListWidget )

        self .includeSubfoldersCheckbox =QCheckBox ("Include Subfolders")
        layout .addWidget (self .includeSubfoldersCheckbox )

        plotButton =QPushButton ("Plot")
        plotButton .clicked .connect (self .plotSelectedFiles )
        layout .addWidget (plotButton )


        self .plotOptions ={
        'Histogram of dI':QCheckBox ('Histogram of dI'),
        'Histogram of dt':QCheckBox ('Histogram of dt'),
        'Histogram of Area':QCheckBox ('Histogram of Area'),
        'Scatter plot dI vs dt':QCheckBox ('Scatter plot dI vs dt'),
        'Scatter plot area vs dt':QCheckBox ('Scatter plot area vs dt'),
        'Histogram of dt (ln)':QCheckBox ('Histogram of dt (ln)'),
        'Scatter plot dI vs dt (ln)':QCheckBox ('Scatter plot dI vs dt (ln)'),
        'Frequency plot':QCheckBox ('Frequency plot')
        }
        for option in self .plotOptions .values ():
            layout .addWidget (option )

        syncZoomButton =QPushButton ("Sync Zoom")
        syncZoomButton .clicked .connect (self .syncZoomAcrossPlots )
        layout .addWidget (syncZoomButton )
        self .selectedFilesOrder =[]


        self .fileListWidget .itemSelectionChanged .connect (self .updateSelectionOrder )


    def setupPlotOptionsTab (self ,tab ):
        layout =QVBoxLayout (tab )


        formLayout =QFormLayout ()


        self .histogramBinsSpinBox =QSpinBox ()
        self .histogramBinsSpinBox .setMinimum (1 )
        self .histogramBinsSpinBox .setMaximum (1000 )
        self .histogramBinsSpinBox .setValue (100 )
        formLayout .addRow ("Histogram Bins:",self .histogramBinsSpinBox )


        self .markerSizeSpinBox =QSpinBox ()
        self .markerSizeSpinBox .setMinimum (1 )
        self .markerSizeSpinBox .setMaximum (100 )
        self .markerSizeSpinBox .setValue (3 )
        formLayout .addRow ("Marker Size:",self .markerSizeSpinBox )


        self .linearFitCheckBox =QCheckBox ("Enable Linear Fit in the Frequency Plots")
        formLayout .addRow (self .linearFitCheckBox )


        self .fitBetweenPointsCheckBox =QCheckBox ("Fit Between Two Points")
        formLayout .addRow (self .fitBetweenPointsCheckBox )


        self .fitStartPointSpinBox =QDoubleSpinBox ()
        self .fitStartPointSpinBox .setMinimum (-np .inf )
        self .fitStartPointSpinBox .setMaximum (np .inf )
        self .fitStartPointSpinBox .setValue (0 )


        self .fitEndPointSpinBox =QDoubleSpinBox ()
        self .fitEndPointSpinBox .setMinimum (-np .inf )
        self .fitEndPointSpinBox .setMaximum (np .inf )
        self .fitEndPointSpinBox .setValue (1 )

        self .fitRangeLabel =QLabel ("Fit Range:")
        self .fitRangeLayout =QHBoxLayout ()
        self .fitRangeLayout .addWidget (self .fitRangeLabel )
        self .fitRangeLayout .addWidget (self .fitStartPointSpinBox )
        self .fitRangeLayout .addWidget (self .fitEndPointSpinBox )
        formLayout .addRow (self .fitRangeLayout )

        self .fitBetweenPointsCheckBox .stateChanged .connect (self .toggleFitRangeVisibility )
        self .toggleFitRangeVisibility (False )

        self .overlayPlotsCheckBox =QCheckBox ("Overlay Plots")
        formLayout .addRow (self .overlayPlotsCheckBox )

        self .alphaSpin =QDoubleSpinBox ()
        self .alphaSpin .setMinimum (0.0 )
        self .alphaSpin .setMaximum (1.0 )
        self .alphaSpin .setSingleStep (0.1 )
        self .alphaSpin .setValue (0.5 )
        formLayout .addRow ("Alpha:",self .alphaSpin )


        self .saveDirectoryLineEdit =QLineEdit ()
        self .saveDirectoryButton =QPushButton ("Select Save Folder")
        self .saveDirectoryButton .clicked .connect (self .selectSaveFolder )
        saveFolderLayout =QHBoxLayout ()
        saveFolderLayout .addWidget (self .saveDirectoryLineEdit )
        saveFolderLayout .addWidget (self .saveDirectoryButton )
        formLayout .addRow ("Save Figures To:",saveFolderLayout )


        updatePlotsButton =QPushButton ("Update Plots")
        updatePlotsButton .clicked .connect (self .updatePlots )
        formLayout .addRow ("",updatePlotsButton )


        saveFiguresButton =QPushButton ("Save Figures")
        saveFiguresButton .clicked .connect (self .saveAllFigures )
        formLayout .addRow ("",saveFiguresButton )

        layout .addLayout (formLayout )


        layout .setSpacing (10 )
        tab .setLayout (layout )

    def setupDatabaseTab (self ,tab ):
        layout =QVBoxLayout (tab )

        self .filterLayout =QVBoxLayout ()

        filterScrollArea =QScrollArea ()
        filterScrollArea .setWidgetResizable (True )
        filterScrollAreaContent =QWidget ()
        filterScrollAreaContent .setLayout (self .filterLayout )
        filterScrollArea .setWidget (filterScrollAreaContent )
        layout .addWidget (filterScrollArea )

        addFilterButton =QPushButton ("Add Filter")
        addFilterButton .clicked .connect (self .addFilter )
        layout .addWidget (addFilterButton )

        buttonLayout =QHBoxLayout ()
        applyFiltersButton =QPushButton ("Apply Filters")
        applyFiltersButton .clicked .connect (self .applyFilters )
        buttonLayout .addWidget (applyFiltersButton )

        saveFilteredDataButton =QPushButton ("Save Filtered Data")
        saveFilteredDataButton .clicked .connect (self .saveFilteredData )
        buttonLayout .addWidget (saveFilteredDataButton )

        layout .addLayout (buttonLayout )

    def addFilter (self ):
        filterWidget =QWidget ()
        filterLayout =QFormLayout (filterWidget )

        variableComboBox =QComboBox ()
        variables =["ΔI","Δt_fwhm","ΔI_fwhm","Area","Δt","Skew","Kurtosis","Event_Baseline","Event_Time"]
        variableComboBox .addItems (variables )
        filterLayout .addRow ("Variable:",variableComboBox )

        minValueSpinBox =QDoubleSpinBox ()
        minValueSpinBox .setMinimum (-np .inf )
        minValueSpinBox .setMaximum (np .inf )
        filterLayout .addRow ("Min Value:",minValueSpinBox )

        maxValueSpinBox =QDoubleSpinBox ()
        maxValueSpinBox .setMinimum (-np .inf )
        maxValueSpinBox .setMaximum (np .inf )
        filterLayout .addRow ("Max Value:",maxValueSpinBox )

        removeFilterButton =QPushButton ("Remove")
        removeFilterButton .clicked .connect (lambda :self .removeFilter (filterWidget ))
        filterLayout .addRow ("Remove:",removeFilterButton )

        self .filterLayout .addWidget (filterWidget )
        self .updateLogicComboBoxes ()


    def updateLogicComboBoxes (self ):
        filterCount =self .filterLayout .count ()

        for i in range (filterCount -1 ):
            filterWidget =self .filterLayout .itemAt (i ).widget ()
            filterLayout =filterWidget .layout ()

            if filterLayout .rowCount ()==4 :
                logicComboBox =QComboBox ()
                logicComboBox .addItems (["AND","OR"])
                filterLayout .addRow ("Logic:",logicComboBox )
            elif i <filterCount -2 :
                logicLabel =None 
                for j in range (filterLayout .rowCount ()):
                    label =filterLayout .itemAt (j ,QFormLayout .ItemRole .LabelRole ).widget ()
                    if isinstance (label ,QLabel )and label .text ()=="Logic:":
                        logicLabel =label 
                        break 
                if logicLabel is None :
                    logicComboBox =QComboBox ()
                    logicComboBox .addItems (["AND","OR"])
                    filterLayout .addRow ("Logic:",logicComboBox )
            else :
                for j in range (filterLayout .rowCount ()):
                    label =filterLayout .itemAt (j ,QFormLayout .ItemRole .LabelRole ).widget ()
                    if isinstance (label ,QLabel )and label .text ()=="Logic:":
                        filterLayout .removeRow (j )
                        break 

    def toggleFitRangeVisibility (self ,visible ):
        self .fitRangeLabel .setVisible (visible )
        self .fitStartPointSpinBox .setVisible (visible )
        self .fitEndPointSpinBox .setVisible (visible )


    def selectSaveFolder (self ):
        folder =QFileDialog .getExistingDirectory (self ,"Select Folder")
        if folder :
            self .saveDirectoryLineEdit .setText (folder )

    def saveAllFigures (self ):
        saveFolder =self .saveDirectoryLineEdit .text ().strip ()
        if not saveFolder :
            QMessageBox .information (self ,"Info","Please select a folder to save the figures.")
            return 

        for i in range (self .rightPanel .count ()):
            tab =self .rightPanel .widget (i )
            if hasattr (tab ,"property"):
                tabCanvases =tab .property ("canvases")
                if tabCanvases :
                    tabFolder =os .path .join (saveFolder ,self .rightPanel .tabText (i ))
                    if not os .path .exists (tabFolder ):
                        os .makedirs (tabFolder )
                    for j ,canvas in enumerate (tabCanvases ):
                        figure =canvas .figure 
                        figure .savefig (os .path .join (tabFolder ,f"figure_{j +1 }.png"),dpi =300 ,bbox_inches ='tight')
        QMessageBox .information (self ,"Info","Figures saved successfully.")

    def removeFilter (self ,filterWidget ):
        self .filterLayout .removeWidget (filterWidget )
        filterWidget .deleteLater ()
        self .updateLogicComboBoxes ()

    def applyFilters (self ):
        self .filtered_data ={}
        selectedFiles =[self .fileListWidget .item (i ).text ()for i in range (self .fileListWidget .count ())if self .fileListWidget .item (i ).isSelected ()]

        for filePath in selectedFiles :
            fileData =np .load (filePath )
            X =fileData ['X']

            mask =np .ones (len (X ),dtype =bool )

            for i in range (self .filterLayout .count ()):
                filterWidget =self .filterLayout .itemAt (i ).widget ()
                filterLayout =filterWidget .layout ()

                for j in range (filterLayout .rowCount ()):
                    label =filterLayout .itemAt (j ,QFormLayout .ItemRole .LabelRole ).widget ()
                    if isinstance (label ,QLabel ):
                        if label .text ()=="Variable:":
                            variableComboBox =filterLayout .itemAt (j ,QFormLayout .ItemRole .FieldRole ).widget ()
                        elif label .text ()=="Min Value:":
                            minValueSpinBox =filterLayout .itemAt (j ,QFormLayout .ItemRole .FieldRole ).widget ()
                        elif label .text ()=="Max Value:":
                            maxValueSpinBox =filterLayout .itemAt (j ,QFormLayout .ItemRole .FieldRole ).widget ()
                        elif label .text ()=="Logic:":
                            logicComboBox =filterLayout .itemAt (j ,QFormLayout .ItemRole .FieldRole ).widget ()

                variable =variableComboBox .currentText ()
                minValue =minValueSpinBox .value ()
                maxValue =maxValueSpinBox .value ()

                variableIndex =["ΔI","Δt_fwhm","ΔI_fwhm","Area","Δt","Skew","Kurtosis","Event_Baseline","Event_Time"].index (variable )
                variableValues =X [:,variableIndex ]

                condition =(variableValues >=minValue )&(variableValues <=maxValue )

                if i ==0 :
                    mask =condition 
                else :
                    if isinstance (logicComboBox ,QComboBox ):
                        logic =logicComboBox .currentText ()
                        if logic =="AND":
                            mask &=condition 
                        else :
                            mask |=condition 
                    else :
                        mask &=condition 

            self .filtered_data [filePath ]=X [mask ]

        self .plotSelectedFiles ()

    def selectFolder (self ):
        folderPath =QFileDialog .getExistingDirectory (self ,"Select Folder")
        if folderPath :
            self .populateFileList (folderPath ,self .includeSubfoldersCheckbox .isChecked ())

    def saveFilteredData (self ):
        saveFolder =QFileDialog .getExistingDirectory (self ,"Select Folder to Save Filtered Data")
        if saveFolder :
            for filePath ,filteredData in self .filtered_data .items ():
                fileName =os .path .basename (filePath )
                savePath =os .path .join (saveFolder ,f"filtered_{fileName }")


                fileData =np .load (filePath )
                settings =fileData ['settings'].item ()


                np .savez (savePath ,X =filteredData ,settings =settings )

            QMessageBox .information (self ,"Info","Filtered data and settings saved successfully.")

    def populateFileList (self ,folderPath ,includeSubfolders ,sort =True ):
        self .fileListWidget .clear ()
        file_list =[]

        if includeSubfolders :
            for root ,dirs ,files in os .walk (folderPath ):
                for file in files :
                    if file .endswith ('.dataset.npz'):
                        fullPath =os .path .join (root ,file )
                        file_list .append (fullPath )
        else :
            for file in os .listdir (folderPath ):
                if file .endswith ('.dataset.npz'):
                    fullPath =os .path .join (folderPath ,file )
                    file_list .append (fullPath )

        if sort :
            file_list .sort (key =lambda x :os .path .basename (x ).lower ())

        for file in file_list :
            self .fileListWidget .addItem (file )















    def updateSelectionOrder (self ):

        currentSelection ={self .fileListWidget .item (i ).text ()for i in range (self .fileListWidget .count ())if self .fileListWidget .item (i ).isSelected ()}


        self .selectedFilesOrder =[f for f in self .selectedFilesOrder if f in currentSelection ]


        for i in range (self .fileListWidget .count ()):
            itemText =self .fileListWidget .item (i ).text ()
            if self .fileListWidget .item (i ).isSelected ()and itemText not in self .selectedFilesOrder :
                self .selectedFilesOrder .append (itemText )


    def updatePlots (self ):

        self .histogram_bins =self .histogramBinsSpinBox .value ()
        self .scatter_marker_size =self .markerSizeSpinBox .value ()


        self .plotSelectedFiles ()


    @staticmethod 
    def create_joint_plot (self ,fig ,data_x ,data_y ,min_x ,max_x ,min_y ,max_y ,plotType ,alpha ):

        gs =fig .add_gridspec (4 ,4 )


        ax_scatter =fig .add_subplot (gs [1 :4 ,0 :3 ])
        if data_x is not None and data_y is not None :
            ax_scatter .scatter (data_x ,data_y ,s =int (self .markerSizeSpinBox .value ()),alpha =alpha )
        ax_scatter .set_xlim (min_x ,max_x )
        ax_scatter .set_ylim (min_y ,max_y )


        ax_histx =fig .add_subplot (gs [0 ,0 :3 ],sharex =ax_scatter )
        if data_x is not None :
            ax_histx .hist (data_x ,bins =self .histogramBinsSpinBox .value ())
        ax_histx .set_ylabel (' ')


        ax_histy =fig .add_subplot (gs [1 :4 ,3 ],sharey =ax_scatter )
        if data_y is not None :
            ax_histy .hist (data_y ,bins =self .histogramBinsSpinBox .value (),orientation ='horizontal')
        ax_histy .set_xlabel (' ')


        plt .setp (ax_histx .get_xticklabels (),visible =False )
        plt .setp (ax_histy .get_yticklabels (),visible =False )

        ax_scatter .xaxis .set_major_formatter (ScalarFormatter (useMathText =True ))
        ax_scatter .yaxis .set_major_formatter (ScalarFormatter (useMathText =True ))
        if 'dI vs dt'in plotType and 'ln'not in plotType :
            ax_scatter .set_xlabel ('Δt (ms)')
            ax_scatter .set_ylabel ('ΔI (nA)')
        elif 'area vs dt'in plotType :
            ax_scatter .set_xlabel ('Δt (ms)')
            ax_scatter .set_ylabel ('Area (C)')
        elif 'dI vs dt (ln)'in plotType :
            ax_scatter .set_xlabel ('ln(Δt (ms))')
            ax_scatter .set_ylabel ('ΔI (nA)')

        ax_scatter .ticklabel_format (style ='sci',axis ='both',scilimits =(0 ,0 ))
        ax_histx .ticklabel_format (style ='sci',axis ='both',scilimits =(0 ,0 ))
        ax_histy .ticklabel_format (style ='sci',axis ='both',scilimits =(0 ,0 ))
        fig .tight_layout ()


        return {'ax_scatter':ax_scatter ,'ax_histx':ax_histx ,'ax_histy':ax_histy }

    def plotSelectedFiles (self ):
        '''selectedFiles = [self.fileListWidget.item(i).text() for i in range(self.fileListWidget.count()) if self.fileListWidget.item(i).isSelected()]
        plotOptions = {key: option.isChecked() for key, option in self.plotOptions.items() if option.isChecked()}'''

        plotOptions ={key :option .isChecked ()for key ,option in self .plotOptions .items ()if option .isChecked ()}
        self .rightPanel .clear ()

        self .rightPanel .clear ()

        if not self .selectedFilesOrder or not plotOptions :
            return 

        data =self .prepareData (self .selectedFilesOrder ,plotOptions )

        for plotType ,plotData in data .items ():
            tab =QWidget ()
            layout =QVBoxLayout (tab )
            plotLayout =QGridLayout ()
            alpha =self .alphaSpin .value ()
            tabCanvases =[]

            if self .overlayPlotsCheckBox .isChecked ():

                fig =Figure (figsize =(10 ,5 ))
                canvas =FigureCanvas (fig )
                tabCanvases .append (canvas )

                min_x ,max_x ,min_y ,max_y =self .calculateGlobalAxisLimits (plotData ,plotType )

                if 'Scatter plot dI vs dt'in plotType and 'ln'not in plotType :

                    axes_dict =self .create_joint_plot (self ,fig ,None ,None ,min_x ,max_x ,min_y ,max_y ,'dI vs dt',alpha )
                    ax_scatter =axes_dict ['ax_scatter']
                    ax_histx =axes_dict ['ax_histx']
                    ax_histy =axes_dict ['ax_histy']

                    for filePath ,values in plotData .items ():
                        alpha =self .alphaSpin .value ()
                        label =os .path .basename (filePath )

                        ax_scatter .scatter (values [0 ],values [1 ],s =int (self .markerSizeSpinBox .value ()),alpha =alpha ,label =label )
                        ax_histx .hist (values [0 ],bins =self .histogramBinsSpinBox .value (),alpha =alpha )
                        ax_histy .hist (values [1 ],bins =self .histogramBinsSpinBox .value (),orientation ='horizontal',alpha =alpha )

                    ax_scatter .set_xlabel ('Δt (ms)')
                    ax_scatter .set_ylabel ('ΔI (nA)')
                    ax_histx .set_ylabel ('Count')
                    ax_histy .set_xlabel ('Count')
                    fig .tight_layout ()
                    ax_scatter .legend ()

                elif 'Scatter plot area vs dt'in plotType :

                    axes_dict =self .create_joint_plot (self ,fig ,None ,None ,min_x ,max_x ,min_y ,max_y ,'area vs dt',alpha )
                    ax_scatter =axes_dict ['ax_scatter']
                    ax_histx =axes_dict ['ax_histx']
                    ax_histy =axes_dict ['ax_histy']

                    for filePath ,values in plotData .items ():
                        alpha =self .alphaSpin .value ()
                        label =os .path .basename (filePath )

                        ax_scatter .scatter (values [0 ],values [1 ],s =int (self .markerSizeSpinBox .value ()),alpha =alpha ,label =label )
                        ax_histx .hist (values [0 ],bins =self .histogramBinsSpinBox .value (),alpha =alpha )
                        ax_histy .hist (values [1 ],bins =self .histogramBinsSpinBox .value (),orientation ='horizontal',alpha =alpha )
                    ax_scatter .set_xlabel ('Δt (ms)')
                    ax_scatter .set_ylabel ('Area')
                    ax_histx .set_ylabel ('Count')
                    ax_histy .set_xlabel ('Count')
                    fig .tight_layout ()
                    ax_scatter .legend ()

                elif 'Scatter plot dI vs dt (ln)'in plotType :

                    axes_dict =self .create_joint_plot (self ,fig ,None ,None ,np .log (min_x ),np .log (max_x ),min_y ,max_y ,'dI vs dt (ln)',alpha )
                    ax_scatter =axes_dict ['ax_scatter']
                    ax_histx =axes_dict ['ax_histx']
                    ax_histy =axes_dict ['ax_histy']

                    for filePath ,values in plotData .items ():
                        alpha =self .alphaSpin .value ()
                        label =os .path .basename (filePath )

                        ax_scatter .scatter (np .log (values [0 ]),values [1 ],s =int (self .markerSizeSpinBox .value ()),alpha =alpha ,label =label )
                        ax_histx .hist (np .log (values [0 ]),bins =self .histogramBinsSpinBox .value (),alpha =alpha )
                        ax_histy .hist (values [1 ],bins =self .histogramBinsSpinBox .value (),orientation ='horizontal',alpha =alpha )
                    ax_scatter .set_xlabel ('ln(Δt (ms))')
                    ax_scatter .set_ylabel ('ΔI (nA)')
                    ax_histx .set_ylabel ('Count')
                    ax_histy .set_xlabel ('Count')
                    fig .tight_layout ()
                    ax_scatter .legend ()


                elif 'Histogram of dt (ln)'in plotType :

                    ax =fig .add_subplot (111 )

                    for filePath ,values in plotData .items ():
                        alpha =self .alphaSpin .value ()
                        label =os .path .basename (filePath )
                        log_values =np .log (values [values >0 ]*1e3 )

                        ax .hist (log_values ,bins =self .histogramBinsSpinBox .value (),range =(np .log (min_x *1e3 ),np .log (max_x *1e3 )),alpha =alpha ,label =label ,density =True )
                    ax .set_xlabel ('ln(Δt (ms))')
                    ax .set_ylabel ('Density')
                    fig .tight_layout ()
                    ax .legend ()

                elif 'Frequency plot'in plotType :

                    ax =fig .add_subplot (111 )

                    for filePath ,values in plotData .items ():
                        alpha =self .alphaSpin .value ()
                        label =os .path .basename (filePath )

                        ax .plot (values [0 ],values [1 ],'o',markersize =self .scatter_marker_size ,alpha =alpha ,label =label )

                        if self .linearFitCheckBox .isChecked ():
                            if self .fitBetweenPointsCheckBox .isChecked ():
                                start_point =self .fitStartPointSpinBox .value ()
                                end_point =self .fitEndPointSpinBox .value ()
                                mask =(values [0 ]>=start_point )&(values [0 ]<=end_point )
                                x_fit =values [0 ][mask ]
                                y_fit =values [1 ][mask ]
                                slope ,intercept =np .polyfit (x_fit ,y_fit ,1 )
                            else :
                                slope ,intercept =np .polyfit (values [0 ],values [1 ],1 )
                                x_fit =np .array ([min_x ,max_x ])
                                y_fit =slope *x_fit +intercept 

                            ax .plot (x_fit ,slope *x_fit +intercept ,'--',alpha =alpha )

                    ax .legend ()
                    ax .set_xlabel ('Time (s)')
                    ax .set_ylabel ('Event Number')



                else :

                    ax =fig .add_subplot (111 )
                    for filePath ,values in plotData .items ():
                        alpha =self .alphaSpin .value ()
                        label =os .path .basename (filePath )
                        ax .hist (values ,bins =self .histogramBinsSpinBox .value (),range =(min_x ,max_x ),alpha =alpha ,label =label ,density =True )

                    if 'Area'in plotType :
                        ax .set_xlabel ('Area')
                    elif 'dt'in plotType and 'ln'not in plotType :
                        ax .set_xlabel ('Δt (s)')
                    elif 'dI'in plotType :
                        ax .set_xlabel ('ΔI (nA)')
                    elif 'dt (ln)'in plotType :
                        ax .set_xlabel ('ln(Δt (ms))')

                    ax .set_ylabel ('Density')
                    ax .legend ()

                fig .suptitle ("Overlaid Plots",fontsize =12 )

                fig .tight_layout ()
                fig .tight_layout ()
                fig .tight_layout ()
                fig .tight_layout ()
                fig .tight_layout ()
                plotLayout .addWidget (canvas ,0 ,0 )



            else :
                row =col =0 
                max_col =1 

                min_x ,max_x ,min_y ,max_y =self .calculateGlobalAxisLimits (plotData ,plotType )

                for filePath ,values in plotData .items ():
                    fig =Figure (figsize =(5 ,5 ))
                    canvas =FigureCanvas (fig )

                    if 'Scatter plot dI vs dt'in plotType and 'ln'not in plotType :
                        axes_dict =self .create_joint_plot (self ,fig ,values [0 ],values [1 ],min_x ,max_x ,min_y ,max_y ,'dI vs dt',alpha =1.0 )
                        canvas .axes_dict =axes_dict 
                        canvas .plotType ='Scatter plot dI vs dt'
                        title =f"{os .path .basename (filePath )}\nEvents: {len (values [0 ])}"

                    elif 'Scatter plot area vs dt'in plotType :
                        axes_dict =self .create_joint_plot (self ,fig ,values [0 ],values [1 ],min_x ,max_x ,min_y ,max_y ,'area vs dt',alpha =1.0 )
                        canvas .axes_dict =axes_dict 
                        canvas .plotType ='Scatter plot area vs dt'
                        title =f"{os .path .basename (filePath )}\nEvents: {len (values [0 ])}"

                    elif 'Scatter plot dI vs dt (ln)'in plotType :
                        axes_dict =self .create_joint_plot (self ,fig ,np .log (values [0 ]),values [1 ],np .log (min_x ),np .log (max_x ),min_y ,max_y ,'dI vs dt (ln)',alpha =1.0 )
                        canvas .axes_dict =axes_dict 
                        canvas .plotType ='Scatter plot dI vs dt (ln)'
                        title =f"{os .path .basename (filePath )}\nEvents: {len (values [0 ])}"

                    elif 'Histogram of dt (ln)'in plotType :
                        log_values =np .log (values [values >0 ]*1e3 )
                        axes_dict =self .create_histogram (fig ,log_values ,np .log (min_x *1e3 ),np .log (max_x *1e3 ),'dt (ln)',alpha )
                        canvas .axes_dict =axes_dict 
                        canvas .plotType ='Histogram of dt (ln)'
                        title =f"{os .path .basename (filePath )}\nEvents: {len (log_values )}"

                    elif 'Frequency plot'in plotType :
                        self .create_frequency_plot (fig ,values [0 ],values [1 ],min_x ,max_x ,min_y ,max_y ,alpha =1.0 )
                        canvas .plotType ='Frequency plot'
                        title =f"{os .path .basename (filePath )}\nEvents: {len (values [0 ])}"
                        if self .linearFitCheckBox .isChecked ():
                            if self .fitBetweenPointsCheckBox .isChecked ():
                                start_point =self .fitStartPointSpinBox .value ()
                                end_point =self .fitEndPointSpinBox .value ()
                                mask =(values [0 ]>=start_point )&(values [0 ]<=end_point )
                                x_fit =values [0 ][mask ]
                                y_fit =values [1 ][mask ]
                                slope ,intercept =np .polyfit (x_fit ,y_fit ,1 )
                            else :
                                slope ,intercept =np .polyfit (values [0 ],values [1 ],1 )
                                x_fit =np .array ([min_x ,max_x ])
                                y_fit =slope *x_fit +intercept 

                            ax =fig .axes [0 ]
                            ax .plot (x_fit ,slope *x_fit +intercept ,'r--',label =f'Fit: {slope :.2f} Events/s, Intercept: {intercept :.2f}')
                            ax .legend ()
                            title +=f"\nFit: {slope :.2f} Events/s, Intercept: {intercept :.2f}"

                    else :

                        axes_dict =self .create_histogram (fig ,values ,min_x ,max_x ,plotType ,alpha =1.0 )
                        canvas .axes_dict =axes_dict 
                        canvas .plotType =plotType 
                        title =f"{os .path .basename (filePath )}\nEvents: {len (values )}"

                    fig .suptitle (title ,fontsize =8 )

                    plotLayout .addWidget (canvas ,row ,col )
                    tabCanvases .append (canvas )

                    col +=1 
                    if col >max_col :
                        col =0 
                        row +=1 


                    fig .tight_layout ()

            layout .addLayout (plotLayout )

            if tabCanvases :
                toolbar =NavigationToolbar (tabCanvases [0 ],self )
                layout .addWidget (toolbar )

            fig .tight_layout ()

            resetButton =QPushButton ("Reset Zoom")
            resetButton .clicked .connect (self .resetZoom )
            layout .addWidget (resetButton )

            self .rightPanel .addTab (tab ,plotType )
            tab .setProperty ("canvases",tabCanvases )


    def create_frequency_plot (self ,fig ,times ,event_numbers ,min_x ,max_x ,min_y ,max_y ,alpha ):
        ax =fig .add_subplot (111 )
        ax .plot (times ,event_numbers ,'o',markersize =self .scatter_marker_size ,alpha =alpha )
        ax .set_xlim (min_x ,max_x )

        ax .set_xlabel ('Time (s)')
        ax .set_ylabel ('Event Number')
        ax .ticklabel_format (style ='sci',axis ='both',scilimits =(0 ,0 ))
        fig .tight_layout ()


    def syncZoomAcrossPlots (self ):
        currentTab =self .rightPanel .currentWidget ()
        if hasattr (currentTab ,"property"):
            tabCanvases =currentTab .property ("canvases")
            if tabCanvases and len (tabCanvases )>0 :

                referenceAxes =tabCanvases [0 ].figure .axes [0 ]
                xlim =referenceAxes .get_xlim ()
                ylim =referenceAxes .get_ylim ()

                for canvas in tabCanvases :
                    plotType =getattr (canvas ,'plotType','')
                    if 'Scatter plot'in plotType :
                        if hasattr (canvas ,'axes_dict'):
                            canvas .axes_dict ['ax_scatter'].set_xlim (xlim )
                            canvas .axes_dict ['ax_scatter'].set_ylim (ylim )
                            canvas .axes_dict ['ax_histx'].set_xlim (xlim )
                            canvas .axes_dict ['ax_histy'].set_ylim (ylim )
                    elif 'Histogram'in plotType or 'Frequency plot'in plotType :

                        ax =canvas .figure .axes [0 ]
                        ax .set_xlim (xlim )
                    else :

                        for ax in canvas .figure .axes :
                            ax .set_xlim (xlim )
                            ax .set_ylim (ylim )

                    canvas .draw_idle ()



    def create_histogram (self ,fig ,values ,min_x ,max_x ,plotType ,alpha ):
        ax =fig .add_subplot (111 )
        ax .hist (values ,bins =self .histogramBinsSpinBox .value (),range =(min_x ,max_x ),alpha =alpha )
        ax .xaxis .set_major_formatter (ScalarFormatter (useMathText =True ))
        if 'dI'in plotType :
            ax .set_xlabel ('ΔI (nA)')
        elif 'dt'in plotType and 'ln'not in plotType :
            ax .set_xlabel ('Δt (s)')
        elif 'dt (ln)'in plotType :
            ax .set_xlabel ('ln(Δt (ms))')
        elif 'Area'in plotType :
            ax .set_xlabel ('Area')
        ax .set_ylabel ('Count')
        ax .ticklabel_format (style ='sci',axis ='both',scilimits =(0 ,0 ))
        fig .tight_layout ()


        return {'ax_hist':ax }


    def resetZoom (self ):
        currentTab =self .rightPanel .currentWidget ()
        if currentTab is None :
            return 

        plotType =self .rightPanel .tabText (self .rightPanel .indexOf (currentTab ))

        if plotType in self .global_limits :
            min_x ,max_x ,min_y ,max_y =self .global_limits [plotType ]

            tabCanvases =currentTab .property ("canvases")
            if tabCanvases :
                for canvas in tabCanvases :
                    if plotType =='Scatter plot dI vs dt'or plotType =='Scatter plot area vs dt'or plotType =='Scatter plot dI vs dt (ln)':
                        if hasattr (canvas ,'axes_dict'):
                            canvas .axes_dict ['ax_scatter'].set_xlim (min_x ,max_x )
                            canvas .axes_dict ['ax_scatter'].set_ylim (min_y ,max_y )
                            canvas .axes_dict ['ax_histx'].set_xlim (min_x ,max_x )
                            canvas .axes_dict ['ax_histy'].set_ylim (min_y ,max_y )
                    elif plotType =='Frequency plot':
                        ax =canvas .figure .axes [0 ]
                        ax .set_xlim (min_x ,max_x )

                    elif plotType in ['Histogram of dI','Histogram of dt','Histogram of Area','Histogram of dt (ln)']:
                        ax =canvas .figure .axes [0 ]
                        ax .set_xlim (min_x ,max_x )
                    else :
                        print (f"Unsupported plot type: {plotType }")

                    canvas .draw_idle ()


    def prepareData (self ,selectedFiles ,plotOptions ):
        data ={}
        for filePath in selectedFiles :
            if filePath in self .filtered_data :
                X =self .filtered_data [filePath ]
                fileData ={'X':X }
            else :
                fileData =np .load (filePath )
                X =fileData ['X']

            for option in plotOptions :
                if option not in data :
                    data [option ]={}

                if 'Histogram'in option :
                    index =0 if 'dI'in option else 3 if 'Area'in option else 4 
                    values =X [:,index ].astype (float )

                    values =values [~np .isnan (values )]
                    data [option ][filePath ]=values 
                elif 'Histogram of dt (ln)'in option :

                    dt_values =X [:,4 ]

                    positive_dt_values =dt_values [dt_values >0 ]

                    log_dt_values =np .log (positive_dt_values *1e3 )

                    log_dt_values =log_dt_values [~np .isnan (log_dt_values )]
                    data ['Histogram of dt (ln)'][filePath ]=log_dt_values .astype (float )
                elif 'Frequency plot'in option :
                    event_numbers =np .arange (1 ,len (X )+1 ).astype (float )
                    times =X [:,8 ].astype (float )


                    valid_indices =~np .isnan (times )
                    event_numbers =event_numbers [valid_indices ]
                    times =times [valid_indices ]
                    data [option ][filePath ]=(times ,event_numbers )
                else :
                    x_index ,y_index =(4 ,0 )if 'dI vs dt'in option else (4 ,3 )
                    x_values =X [:,x_index ]*1e3 
                    y_values =X [:,y_index ]

                    valid_indices =~np .isnan (x_values )&~np .isnan (y_values )
                    x_values =x_values [valid_indices ]
                    y_values =y_values [valid_indices ]
                    data [option ][filePath ]=(x_values ,y_values )

        return data 

    def calculateGlobalAxisLimits (self ,plotData ,plotType ):
        min_x =min_y =np .inf 
        max_x =max_y =-np .inf 
        for values in plotData .values ():
            if isinstance (values ,dict ):

                for subvalues in values .values ():
                    if 'Histogram'in plotType :
                        curr_min_x =np .min (subvalues )
                        curr_max_x =np .max (subvalues )

                        if not np .isfinite (curr_min_x )or not np .isfinite (curr_max_x ):
                            continue 
                        if curr_min_x <curr_max_x :
                            min_x =min (min_x ,curr_min_x )
                            max_x =max (max_x ,curr_max_x )
                        else :
                            min_x =min (min_x ,curr_min_x )
                            max_x =max (max_x ,curr_min_x +1e-6 )
                    elif 'Frequency plot'in plotType :
                        x ,y =subvalues 
                        min_x =min (min_x ,np .min (x ))
                        max_x =max (max_x ,np .max (x ))
                        curr_min_y =np .min (y )
                        curr_max_y =np .max (y )
                        if not np .isfinite (curr_min_y )or not np .isfinite (curr_max_y ):
                            continue 
                        min_y =min (min_y ,curr_min_y )
                        max_y =max (max_y ,curr_max_y )
                    else :
                        x ,y =subvalues 
                        min_x =min (min_x ,np .min (x ))
                        max_x =max (max_x ,np .max (x ))
                        curr_min_y =np .min (y )
                        curr_max_y =np .max (y )
                        if not np .isfinite (curr_min_y )or not np .isfinite (curr_max_y ):
                            continue 
                        min_y =min (min_y ,curr_min_y )
                        max_y =max (max_y ,curr_max_y )
            else :
                if 'Histogram'in plotType :
                    curr_min_x =np .min (values )
                    curr_max_x =np .max (values )

                    if not np .isfinite (curr_min_x )or not np .isfinite (curr_max_x ):
                        continue 
                    if curr_min_x <curr_max_x :
                        min_x =min (min_x ,curr_min_x )
                        max_x =max (max_x ,curr_max_x )
                    else :
                        min_x =min (min_x ,curr_min_x )
                        max_x =max (max_x ,curr_min_x +1e-6 )
                elif 'Frequency plot'in plotType :
                    x ,y =values 
                    min_x =min (min_x ,np .min (x ))
                    max_x =max (max_x ,np .max (x ))
                    curr_min_y =np .min (y )
                    curr_max_y =np .max (y )
                    if not np .isfinite (curr_min_y )or not np .isfinite (curr_max_y ):
                        continue 
                    min_y =min (min_y ,curr_min_y )
                    max_y =max (max_y ,curr_max_y )
                else :
                    x ,y =values 
                    min_x =min (min_x ,np .min (x ))
                    max_x =max (max_x ,np .max (x ))
                    curr_min_y =np .min (y )
                    curr_max_y =np .max (y )
                    if not np .isfinite (curr_min_y )or not np .isfinite (curr_max_y ):
                        continue 
                    min_y =min (min_y ,curr_min_y )
                    max_y =max (max_y ,curr_max_y )

        if not np .isfinite (min_x )or not np .isfinite (max_x ):
            min_x ,max_x =0 ,1 
        if not np .isfinite (min_y )or not np .isfinite (max_y ):
            min_y ,max_y =0 ,1 


        self .global_limits [plotType ]=(min_x ,max_x ,min_y ,max_y )

        return min_x ,max_x ,min_y ,max_y 


if __name__ =="__main__":

    app =QApplication (sys .argv )
    app .setStyle (QStyleFactory .create ('Fusion'))


    palette =QPalette ()
    palette .setColor (QPalette .ColorRole .Window ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .WindowText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Base ,QColor (25 ,25 ,25 ))
    palette .setColor (QPalette .ColorRole .AlternateBase ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .ToolTipBase ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .ToolTipText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Text ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .Button ,QColor (53 ,53 ,53 ))
    palette .setColor (QPalette .ColorRole .ButtonText ,Qt .GlobalColor .white )
    palette .setColor (QPalette .ColorRole .BrightText ,Qt .GlobalColor .red )
    palette .setColor (QPalette .ColorRole .Link ,QColor (42 ,130 ,218 ))
    palette .setColor (QPalette .ColorRole .Highlight ,QColor (42 ,130 ,218 ))
    palette .setColor (QPalette .ColorRole .HighlightedText ,Qt .GlobalColor .black )
    app .setPalette (palette )

    mainWindow =PlottingApplication ()
    mainWindow .showMaximized ()
    sys .exit (app .exec ())



"""
A helper module that can work with paths
that can refer to data inside a zipfile
"""
import errno as _errno 
import os as _os 
import stat as _stat 
import sys as _sys 
import time as _time 
import zipfile as _zipfile 

_DFLT_DIR_MODE =(
_stat .S_IXOTH 
|_stat .S_IXGRP 
|_stat .S_IXUSR 
|_stat .S_IROTH 
|_stat .S_IRGRP 
|_stat .S_IRUSR 
)

_DFLT_FILE_MODE =_stat .S_IROTH |_stat .S_IRGRP |_stat .S_IRUSR 


if _sys .version_info [0 ]==2 :
    from StringIO import StringIO as _BaseBytesIO 
    from StringIO import StringIO as _BaseStringIO 

    class _StringIO (_BaseStringIO ):
        def __enter__ (self ):
            return self 

        def __exit__ (self ,exc_type ,exc_value ,traceback ):
            self .close ()
            return False 

    class _BytesIO (_BaseBytesIO ):
        def __enter__ (self ):
            return self 

        def __exit__ (self ,exc_type ,exc_value ,traceback ):
            self .close ()
            return False 

else :
    from io import BytesIO as _BytesIO 
    from io import StringIO as _StringIO 


def _locate (path ):
    full_path =path 
    if _os .path .exists (path ):
        return path ,None 

    else :
        rest =[]
        root =_os .path .splitdrive (path )
        while path and path !=root :
            path ,bn =_os .path .split (path )
            rest .append (bn )
            if _os .path .exists (path ):
                break 

        if path ==root :
            raise IOError (_errno .ENOENT ,full_path ,"No such file or directory")

        if not _os .path .isfile (path ):
            raise IOError (_errno .ENOENT ,full_path ,"No such file or directory")

        rest .reverse ()
        return path ,"/".join (rest ).strip ("/")


_open =open 


def open (path ,mode ="r"):
    if "w"in mode or "a"in mode :
        raise IOError (_errno .EINVAL ,path ,"Write access not supported")
    elif "r+"in mode :
        raise IOError (_errno .EINVAL ,path ,"Write access not supported")

    full_path =path 
    path ,rest =_locate (path )
    if not rest :
        return _open (path ,mode )

    else :
        try :
            zf =_zipfile .ZipFile (path ,"r")

        except _zipfile .error :
            raise IOError (_errno .ENOENT ,full_path ,"No such file or directory")

        try :
            data =zf .read (rest )
        except (_zipfile .error ,KeyError ):
            zf .close ()
            raise IOError (_errno .ENOENT ,full_path ,"No such file or directory")
        zf .close ()

        if mode =="rb":
            return _BytesIO (data )

        else :
            if _sys .version_info [0 ]==3 :
                data =data .decode ("ascii")

            return _StringIO (data )


def listdir (path ):
    full_path =path 
    path ,rest =_locate (path )
    if not rest and not _os .path .isfile (path ):
        return _os .listdir (path )

    else :
        try :
            zf =_zipfile .ZipFile (path ,"r")

        except _zipfile .error :
            raise IOError (_errno .ENOENT ,full_path ,"No such file or directory")

        result =set ()
        seen =False 
        try :
            for nm in zf .namelist ():
                if rest is None :
                    seen =True 
                    value =nm .split ("/")[0 ]
                    if value :
                        result .add (value )

                elif nm .startswith (rest ):
                    if nm ==rest :
                        seen =True 
                        value =""
                        pass 
                    elif nm [len (rest )]=="/":
                        seen =True 
                        value =nm [len (rest )+1 :].split ("/")[0 ]
                    else :
                        value =None 

                    if value :
                        result .add (value )
        except _zipfile .error :
            zf .close ()
            raise IOError (_errno .ENOENT ,full_path ,"No such file or directory")

        zf .close ()

        if not seen :
            raise IOError (_errno .ENOENT ,full_path ,"No such file or directory")

        return list (result )


def isfile (path ):
    full_path =path 
    path ,rest =_locate (path )
    if not rest :
        ok =_os .path .isfile (path )
        if ok :
            try :
                zf =_zipfile .ZipFile (path ,"r")
                return False 
            except (_zipfile .error ,IOError ):
                return True 
        return False 

    zf =None 
    try :
        zf =_zipfile .ZipFile (path ,"r")
        zf .getinfo (rest )
        zf .close ()
        return True 
    except (KeyError ,_zipfile .error ):
        if zf is not None :
            zf .close ()


        try :
            zf .getinfo (rest +"/")
        except KeyError :
            pass 
        else :
            return False 

        rest =rest +"/"
        for nm in zf .namelist ():
            if nm .startswith (rest ):

                return False 


        raise IOError (_errno .ENOENT ,full_path ,"No such file or directory")


def isdir (path ):
    full_path =path 
    path ,rest =_locate (path )
    if not rest :
        ok =_os .path .isdir (path )
        if not ok :
            try :
                zf =_zipfile .ZipFile (path ,"r")
            except (_zipfile .error ,IOError ):
                return False 
            return True 
        return True 

    zf =None 
    try :
        try :
            zf =_zipfile .ZipFile (path )
        except _zipfile .error :
            raise IOError (_errno .ENOENT ,full_path ,"No such file or directory")

        try :
            zf .getinfo (rest )
        except KeyError :
            pass 
        else :

            return False 

        rest =rest +"/"
        try :
            zf .getinfo (rest )
        except KeyError :
            pass 
        else :

            return True 

        for nm in zf .namelist ():
            if nm .startswith (rest ):
                return True 

        raise IOError (_errno .ENOENT ,full_path ,"No such file or directory")
    finally :
        if zf is not None :
            zf .close ()


def islink (path ):
    full_path =path 
    path ,rest =_locate (path )
    if not rest :
        return _os .path .islink (path )

    try :
        zf =_zipfile .ZipFile (path )
    except _zipfile .error :
        raise IOError (_errno .ENOENT ,full_path ,"No such file or directory")
    try :
        try :
            zf .getinfo (rest )
        except KeyError :
            pass 
        else :

            return False 

        rest +="/"
        try :
            zf .getinfo (rest )
        except KeyError :
            pass 
        else :

            return False 

        for nm in zf .namelist ():
            if nm .startswith (rest ):

                return False 

        raise IOError (_errno .ENOENT ,full_path ,"No such file or directory")

    finally :
        zf .close ()


def readlink (path ):
    full_path =path 
    path ,rest =_locate (path )
    if rest :

        raise OSError (_errno .ENOENT ,full_path ,"No such file or directory")

    return _os .readlink (path )


def getmode (path ):
    full_path =path 
    path ,rest =_locate (path )
    if not rest :
        return _stat .S_IMODE (_os .stat (path ).st_mode )

    zf =None 
    try :
        zf =_zipfile .ZipFile (path )
        info =None 

        try :
            info =zf .getinfo (rest )
        except KeyError :
            pass 

        if info is None :
            try :
                info =zf .getinfo (rest +"/")
            except KeyError :
                pass 

        if info is None :
            rest =rest +"/"
            for nm in zf .namelist ():
                if nm .startswith (rest ):
                    break 
            else :
                raise IOError (_errno .ENOENT ,full_path ,"No such file or directory")


            return _DFLT_DIR_MODE 


        if (info .external_attr >>16 )!=0 :
            return _stat .S_IMODE (info .external_attr >>16 )
        else :
            return _DFLT_FILE_MODE 

    finally :
        if zf is not None :
            zf .close ()


def getmtime (path ):
    full_path =path 
    path ,rest =_locate (path )
    if not rest :
        return _os .path .getmtime (path )

    zf =None 
    try :
        zf =_zipfile .ZipFile (path )
        info =None 

        try :
            info =zf .getinfo (rest )
        except KeyError :
            pass 

        if info is None :
            try :
                info =zf .getinfo (rest +"/")
            except KeyError :
                pass 

        if info is None :
            rest =rest +"/"
            for nm in zf .namelist ():
                if nm .startswith (rest ):
                    break 
            else :
                raise IOError (_errno .ENOENT ,full_path ,"No such file or directory")




            return _os .path .getmtime (path )

        return _time .mktime (info .date_time +(0 ,0 ,-1 ))

    finally :
        if zf is not None :
            zf .close ()
